@@ -25,7 +25,6 @@ import (
 	"testing"
 	"time"
 
-	"github.com/prometheus/common/model"
 	"github.com/prometheus/prometheus/notifier"
 	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/rules"
@@ -173,7 +172,7 @@ func TestFailedStartupExitCode(t *testing.T) {
 
 	prom := exec.Command(promPath, "--config.file="+fakeInputFile)
 	err := prom.Run()
-	testutil.NotOk(t, err, "")
+	testutil.NotOk(t, err)
 
 	if exitError, ok := err.(*exec.ExitError); ok {
 		status := exitError.Sys().(syscall.WaitStatus)
@@ -276,7 +275,7 @@ func TestWALSegmentSizeBounds(t *testing.T) {
 		}
 
 		err = prom.Wait()
-		testutil.NotOk(t, err, "")
+		testutil.NotOk(t, err)
 		if exitError, ok := err.(*exec.ExitError); ok {
 			status := exitError.Sys().(syscall.WaitStatus)
 			testutil.Equals(t, expectedExitStatus, status.ExitStatus())
@@ -285,31 +284,3 @@ func TestWALSegmentSizeBounds(t *testing.T) {
 		}
 	}
 }
-
-func TestChooseRetention(t *testing.T) {
-	retention1, err := model.ParseDuration("20d")
-	testutil.Ok(t, err)
-	retention2, err := model.ParseDuration("30d")
-	testutil.Ok(t, err)
-
-	cases := []struct {
-		oldFlagRetention model.Duration
-		newFlagRetention model.Duration
-
-		chosen model.Duration
-	}{
-		// Both are default (unset flags).
-		{defaultRetentionDuration, defaultRetentionDuration, defaultRetentionDuration},
-		// Old flag is set and new flag is unset.
-		{retention1, defaultRetentionDuration, retention1},
-		// Old flag is unset and new flag is set.
-		{defaultRetentionDuration, retention2, retention2},
-		// Both flags are set.
-		{retention1, retention2, retention2},
-	}
-
-	for _, tc := range cases {
-		retention := chooseRetention(tc.oldFlagRetention, tc.newFlagRetention)
-		testutil.Equals(t, tc.chosen, retention)
-	}
-}
