@@ -20,67 +20,67 @@ import (
 	"unicode/utf8"
 )
 
-// item represents a token or text string returned from the scanner.
-type item struct {
-	typ ItemType // The type of this item.
-	pos Pos      // The starting position, in bytes, of this item in the input string.
-	val string   // The value of this item.
+// Item represents a token or text string returned from the scanner.
+type Item struct {
+	Typ ItemType // The type of this Item.
+	Pos Pos      // The starting position, in bytes, of this Item in the input string.
+	Val string   // The value of this Item.
 }
 
-// String returns a descriptive string for the item.
-func (i item) String() string {
+// String returns a descriptive string for the Item.
+func (i Item) String() string {
 	switch {
-	case i.typ == itemEOF:
+	case i.Typ == EOF:
 		return "EOF"
-	case i.typ == itemError:
-		return i.val
-	case i.typ == itemIdentifier || i.typ == itemMetricIdentifier:
-		return fmt.Sprintf("%q", i.val)
-	case i.typ.isKeyword():
-		return fmt.Sprintf("<%s>", i.val)
-	case i.typ.isOperator():
-		return fmt.Sprintf("<op:%s>", i.val)
-	case i.typ.isAggregator():
-		return fmt.Sprintf("<aggr:%s>", i.val)
-	case len(i.val) > 10:
-		return fmt.Sprintf("%.10q...", i.val)
+	case i.Typ == ERROR:
+		return i.Val
+	case i.Typ == IDENTIFIER || i.Typ == METRIC_IDENTIFIER:
+		return fmt.Sprintf("%q", i.Val)
+	case i.Typ.isKeyword():
+		return fmt.Sprintf("<%s>", i.Val)
+	case i.Typ.isOperator():
+		return fmt.Sprintf("<op:%s>", i.Val)
+	case i.Typ.isAggregator():
+		return fmt.Sprintf("<aggr:%s>", i.Val)
+	case len(i.Val) > 10:
+		return fmt.Sprintf("%.10q...", i.Val)
 	}
-	return fmt.Sprintf("%q", i.val)
+	return fmt.Sprintf("%q", i.Val)
 }
 
-// isOperator returns true if the item corresponds to a arithmetic or set operator.
+// isOperator returns true if the Item corresponds to a arithmetic or set operator.
 // Returns false otherwise.
 func (i ItemType) isOperator() bool { return i > operatorsStart && i < operatorsEnd }
 
-// isAggregator returns true if the item belongs to the aggregator functions.
+// isAggregator returns true if the Item belongs to the aggregator functions.
 // Returns false otherwise
 func (i ItemType) isAggregator() bool { return i > aggregatorsStart && i < aggregatorsEnd }
 
-// isAggregator returns true if the item is an aggregator that takes a parameter.
+// isAggregator returns true if the Item is an aggregator that takes a parameter.
 // Returns false otherwise
 func (i ItemType) isAggregatorWithParam() bool {
-	return i == itemTopK || i == itemBottomK || i == itemCountValues || i == itemQuantile
+	return i == TOPK || i == BOTTOMK || i == COUNT_VALUES || i == QUANTILE
 }
 
-// isKeyword returns true if the item corresponds to a keyword.
+// isKeyword returns true if the Item corresponds to a keyword.
 // Returns false otherwise.
 func (i ItemType) isKeyword() bool { return i > keywordsStart && i < keywordsEnd }
 
-// isCompairsonOperator returns true if the item corresponds to a comparison operator.
+// isComparisonOperator returns true if the Item corresponds to a comparison operator.
 // Returns false otherwise.
 func (i ItemType) isComparisonOperator() bool {
 	switch i {
-	case itemEQL, itemNEQ, itemLTE, itemLSS, itemGTE, itemGTR:
+	case EQL, NEQ, LTE, LSS, GTE, GTR:
 		return true
 	default:
 		return false
 	}
 }
 
-// isSetOperator returns whether the item corresponds to a set operator.
+// isSetOperator returns whether the Item corresponds to a set operator.
 func (i ItemType) isSetOperator() bool {
 	switch i {
-	case itemLAND, itemLOR, itemLUnless:
+	case LAND, LOR, LUNLESS:
 		return true
 	}
 	return false
@@ -94,17 +94,17 @@ const LowestPrec = 0 // Non-operators.
 // is LowestPrec.
 func (i ItemType) precedence() int {
 	switch i {
-	case itemLOR:
+	case LOR:
 		return 1
-	case itemLAND, itemLUnless:
+	case LAND, LUNLESS:
 		return 2
-	case itemEQL, itemNEQ, itemLTE, itemLSS, itemGTE, itemGTR:
+	case EQL, NEQ, LTE, LSS, GTE, GTR:
 		return 3
-	case itemADD, itemSUB:
+	case ADD, SUB:
 		return 4
-	case itemMUL, itemDIV, itemMOD:
+	case MUL, DIV, MOD:
 		return 5
-	case itemPOW:
+	case POW:
 		return 6
 	default:
 		return LowestPrec
@@ -113,7 +113,7 @@ func (i ItemType) precedence() int {
 
 func (i ItemType) isRightAssociative() bool {
 	switch i {
-	case itemPOW:
+	case POW:
 		return true
 	default:
 		return false
@@ -123,185 +123,113 @@ func (i ItemType) isRightAssociative() bool {
 
 type ItemType int
 
-const (
-	itemError ItemType = iota // Error occurred, value is error message
-	itemEOF
-	itemComment
-	itemIdentifier
-	itemMetricIdentifier
-	itemLeftParen
-	itemRightParen
-	itemLeftBrace
-	itemRightBrace
-	itemLeftBracket
-	itemRightBracket
-	itemComma
-	itemAssign
-	itemColon
-	itemSemicolon
-	itemString
-	itemNumber
-	itemDuration
-	itemBlank
-	itemTimes
-	itemSpace
-
-	operatorsStart
-	// Operators.
-	itemSUB
-	itemADD
-	itemMUL
-	itemMOD
-	itemDIV
-	itemLAND
-	itemLOR
-	itemLUnless
-	itemEQL
-	itemNEQ
-	itemLTE
-	itemLSS
-	itemGTE
-	itemGTR
-	itemEQLRegex
-	itemNEQRegex
-	itemPOW
-	operatorsEnd
-
-	aggregatorsStart
-	// Aggregators.
-	itemAvg
-	itemCount
-	itemSum
-	itemMin
-	itemMax
-	itemStddev
-	itemStdvar
-	itemTopK
-	itemBottomK
-	itemCountValues
-	itemQuantile
-	aggregatorsEnd
-
-	keywordsStart
-	// Keywords.
-	itemOffset
-	itemBy
-	itemWithout
-	itemOn
-	itemIgnoring
-	itemGroupLeft
-	itemGroupRight
-	itemBool
-	keywordsEnd
-)
-
 var key = map[string]ItemType{
 	// Operators.
-	"and":    itemLAND,
-	"or":     itemLOR,
-	"unless": itemLUnless,
+	"and":    LAND,
+	"or":     LOR,
+	"unless": LUNLESS,
 
 	// Aggregators.
-	"sum":          itemSum,
-	"avg":          itemAvg,
-	"count":        itemCount,
-	"min":          itemMin,
-	"max":          itemMax,
-	"stddev":       itemStddev,
-	"stdvar":       itemStdvar,
-	"topk":         itemTopK,
-	"bottomk":      itemBottomK,
-	"count_values": itemCountValues,
-	"quantile":     itemQuantile,
+	"sum":          SUM,
+	"avg":          AVG,
+	"count":        COUNT,
+	"min":          MIN,
+	"max":          MAX,
+	"stddev":       STDDEV,
+	"stdvar":       STDVAR,
+	"topk":         TOPK,
+	"bottomk":      BOTTOMK,
+	"count_values": COUNT_VALUES,
+	"quantile":     QUANTILE,
 
 	// Keywords.
-	"offset":      itemOffset,
-	"by":          itemBy,
-	"without":     itemWithout,
-	"on":          itemOn,
-	"ignoring":    itemIgnoring,
-	"group_left":  itemGroupLeft,
-	"group_right": itemGroupRight,
-	"bool":        itemBool,
+	"offset":      OFFSET,
+	"by":          BY,
+	"without":     WITHOUT,
+	"on":          ON,
+	"ignoring":    IGNORING,
+	"group_left":  GROUP_LEFT,
+	"group_right": GROUP_RIGHT,
+	"bool":        BOOL,
 }
 
-// These are the default string representations for common items. It does not
-// imply that those are the only character sequences that can be lexed to such an item.
-var itemTypeStr = map[ItemType]string{
-	itemLeftParen:    "(",
-	itemRightParen:   ")",
-	itemLeftBrace:    "{",
-	itemRightBrace:   "}",
-	itemLeftBracket:  "[",
-	itemRightBracket: "]",
-	itemComma:        ",",
-	itemAssign:       "=",
-	itemColon:        ":",
-	itemSemicolon:    ";",
-	itemBlank:        "_",
-	itemTimes:        "x",
-	itemSpace:        "<space>",
+// These are the default string representations for common Items. It does not
+// imply that those are the only character sequences that can be lexed to such an Item.
+var ItemTypeStr = map[ItemType]string{
+	LEFT_PAREN:    "(",
+	RIGHT_PAREN:   ")",
+	LEFT_BRACE:    "{",
+	RIGHT_BRACE:   "}",
+	LEFT_BRACKET:  "[",
+	RIGHT_BRACKET: "]",
+	COMMA:         ",",
+	ASSIGN:        "=",
+	COLON:         ":",
+	SEMICOLON:     ";",
+	BLANK:         "_",
+	TIMES:         "x",
+	SPACE:         "<space>",
 
-	itemSUB:      "-",
-	itemADD:      "+",
-	itemMUL:      "*",
-	itemMOD:      "%",
-	itemDIV:      "/",
-	itemEQL:      "==",
-	itemNEQ:      "!=",
-	itemLTE:      "<=",
-	itemLSS:      "<",
-	itemGTE:      ">=",
-	itemGTR:      ">",
-	itemEQLRegex: "=~",
-	itemNEQRegex: "!~",
-	itemPOW:      "^",
+	SUB:       "-",
+	ADD:       "+",
+	MUL:       "*",
+	MOD:       "%",
+	DIV:       "/",
+	EQL:       "==",
+	NEQ:       "!=",
+	LTE:       "<=",
+	LSS:       "<",
+	GTE:       ">=",
+	GTR:       ">",
+	EQL_REGEX: "=~",
+	NEQ_REGEX: "!~",
+	POW:       "^",
 }
 
 func init() {
-	// Add keywords to item type strings.
+	// Add keywords to Item type strings.
 	for s, ty := range key {
-		itemTypeStr[ty] = s
+		ItemTypeStr[ty] = s
 	}
 	// Special numbers.
-	key["inf"] = itemNumber
-	key["nan"] = itemNumber
+	key["inf"] = NUMBER
+	key["nan"] = NUMBER
 }
 
 func (i ItemType) String() string {
-	if s, ok := itemTypeStr[i]; ok {
+	if s, ok := ItemTypeStr[i]; ok {
 		return s
 	}
-	return fmt.Sprintf("<item %d>", i)
+	return fmt.Sprintf("<Item %d>", i)
 }
 
-func (i item) desc() string {
-	if _, ok := itemTypeStr[i.typ]; ok {
+func (i Item) desc() string {
+	if _, ok := ItemTypeStr[i.Typ]; ok {
 		return i.String()
 	}
-	if i.typ == itemEOF {
-		return i.typ.desc()
+	if i.Typ == EOF {
+		return i.Typ.desc()
 	}
-	return fmt.Sprintf("%s %s", i.typ.desc(), i)
+	return fmt.Sprintf("%s %s", i.Typ.desc(), i)
 }
 
 func (i ItemType) desc() string {
 	switch i {
-	case itemError:
+	case ERROR:
 		return "error"
-	case itemEOF:
+	case EOF:
 		return "end of input"
-	case itemComment:
+	case COMMENT:
 		return "comment"
-	case itemIdentifier:
+	case IDENTIFIER:
 		return "identifier"
-	case itemMetricIdentifier:
+	case METRIC_IDENTIFIER:
 		return "metric identifier"
-	case itemString:
+	case STRING:
 		return "string"
-	case itemNumber:
+	case NUMBER:
 		return "number"
-	case itemDuration:
+	case DURATION:
 		return "duration"
 	}
 	return fmt.Sprintf("%q", i)
@@ -310,20 +238,20 @@ func (i ItemType) desc() string {
 const eof = -1
 
 // stateFn represents the state of the scanner as a function that returns the next state.
-type stateFn func(*lexer) stateFn
+type stateFn func(*Lexer) stateFn
 
 // Pos is the position in a string.
 type Pos int
 
-// lexer holds the state of the scanner.
-type lexer struct {
-	input   string    // The string being scanned.
-	state   stateFn   // The next lexing function to enter.
-	pos     Pos       // Current position in the input.
-	start   Pos       // Start position of this item.
-	width   Pos       // Width of last rune read from input.
-	lastPos Pos       // Position of most recent item returned by nextItem.
-	items   chan item // Channel of scanned items.
+// Lexer holds the state of the scanner.
+type Lexer struct {
+	input   string  // The string being scanned.
+	state   stateFn // The next lexing function to enter.
+	pos     Pos     // Current position in the input.
+	start   Pos     // Start position of this Item.
+	width   Pos     // Width of last rune read from input.
+	lastPos Pos     // Position of most recent Item returned by NextItem.
+	Items   []Item  // Slice buffer of scanned Items.
 
 	parenDepth  int  // Nesting depth of ( ) exprs.
 	braceOpen   bool // Whether a { is opened.
@@ -337,7 +265,7 @@ type lexer struct {
 }
 
 // next returns the next rune in the input.
-func (l *lexer) next() rune {
+func (l *Lexer) next() rune {
 	if int(l.pos) >= len(l.input) {
 		l.width = 0
 		return eof
@@ -349,30 +277,30 @@ func (l *lexer) next() rune {
 }
 
 // peek returns but does not consume the next rune in the input.
-func (l *lexer) peek() rune {
+func (l *Lexer) peek() rune {
 	r := l.next()
 	l.backup()
 	return r
 }
 
 // backup steps back one rune. Can only be called once per call of next.
-func (l *lexer) backup() {
+func (l *Lexer) backup() {
 	l.pos -= l.width
 }
 
-// emit passes an item back to the client.
-func (l *lexer) emit(t ItemType) {
-	l.items <- item{t, l.start, l.input[l.start:l.pos]}
+// emit passes an Item back to the client.
+func (l *Lexer) emit(t ItemType) {
+	l.Items = append(l.Items, Item{t, l.start, l.input[l.start:l.pos]})
 	l.start = l.pos
 }
 
 // ignore skips over the pending input before this point.
-func (l *lexer) ignore() {
+func (l *Lexer) ignore() {
 	l.start = l.pos
 }
 
 // accept consumes the next rune if it's from the valid set.
-func (l *lexer) accept(valid string) bool {
+func (l *Lexer) accept(valid string) bool {
 	if strings.ContainsRune(valid, l.next()) {
 		return true
 	}
@@ -381,7 +309,7 @@ func (l *lexer) accept(valid string) bool {
 }
 
 // acceptRun consumes a run of runes from the valid set.
-func (l *lexer) acceptRun(valid string) {
+func (l *Lexer) acceptRun(valid string) {
 	for strings.ContainsRune(valid, l.next()) {
 		// consume
 	}
@@ -389,15 +317,15 @@ func (l *lexer) acceptRun(valid string) {
 }
 
 // lineNumber reports which line we're on, based on the position of
-// the previous item returned by nextItem. Doing it this way
+// the previous Item returned by NextItem. Doing it this way
 // means we don't have to worry about peek double counting.
-func (l *lexer) lineNumber() int {
+func (l *Lexer) lineNumber() int {
 	return 1 + strings.Count(l.input[:l.lastPos], "\n")
 }
 
 // linePosition reports at which character in the current line
 // we are on.
-func (l *lexer) linePosition() int {
+func (l *Lexer) linePosition() int {
 	lb := strings.LastIndex(l.input[:l.lastPos], "\n")
 	if lb == -1 {
 		return 1 + int(l.lastPos)
@@ -406,40 +334,46 @@ func (l *lexer) linePosition() int {
 }
 
 // errorf returns an error token and terminates the scan by passing
-// back a nil pointer that will be the next state, terminating l.nextItem.
-func (l *lexer) errorf(format string, args ...interface{}) stateFn {
-	l.items <- item{itemError, l.start, fmt.Sprintf(format, args...)}
+// back a nil pointer that will be the next state, terminating l.NextItem.
+func (l *Lexer) errorf(format string, args ...interface{}) stateFn {
+	l.Items = append(l.Items, Item{ERROR, l.start, fmt.Sprintf(format, args...)})
 	return nil
 }
 
-// nextItem returns the next item from the input.
-func (l *lexer) nextItem() item {
-	item := <-l.items
-	l.lastPos = item.pos
-	return item
+// NextItem returns the next Item from the input.
+func (l *Lexer) NextItem() Item {
+	for len(l.Items) == 0 {
+		if l.state != nil {
+			l.state = l.state(l)
+		} else {
+			l.emit(EOF)
+		}
+	}
+	Item := l.Items[0]
+	l.Items = l.Items[1:]
+	l.lastPos = Item.Pos
+	return Item
 }
 
 // lex creates a new scanner for the input string.
-func lex(input string) *lexer {
-	l := &lexer{
+func Lex(input string) *Lexer {
+	l := &Lexer{
 		input: input,
-		items: make(chan item),
+		state: lexStatements,
 	}
-	go l.run()
 	return l
 }
 
 // run runs the state machine for the lexer.
-func (l *lexer) run() {
+func (l *Lexer) run() {
 	for l.state = lexStatements; l.state != nil; {
 		l.state = l.state(l)
 	}
-	close(l.items)
 }
 
 // Release resources used by lexer.
-func (l *lexer) close() {
-	for range l.items {
+func (l *Lexer) close() {
+	for range l.Items {
 		// Consume.
 	}
 }
@@ -448,7 +382,7 @@ func (l *lexer) close() {
 const lineComment = "#"
 
 // lexStatements is the top-level state for lexing.
-func lexStatements(l *lexer) stateFn {
+func lexStatements(l *Lexer) stateFn {
 	if l.braceOpen {
 		return lexInsideBraces
 	}
@@ -463,52 +397,52 @@ func lexStatements(l *lexer) stateFn {
 		} else if l.bracketOpen {
 			return l.errorf("unclosed left bracket")
 		}
-		l.emit(itemEOF)
+		l.emit(EOF)
 		return nil
 	case r == ',':
-		l.emit(itemComma)
+		l.emit(COMMA)
 	case isSpace(r):
 		return lexSpace
 	case r == '*':
-		l.emit(itemMUL)
+		l.emit(MUL)
 	case r == '/':
-		l.emit(itemDIV)
+		l.emit(DIV)
 	case r == '%':
-		l.emit(itemMOD)
+		l.emit(MOD)
 	case r == '+':
-		l.emit(itemADD)
+		l.emit(ADD)
 	case r == '-':
-		l.emit(itemSUB)
+		l.emit(SUB)
 	case r == '^':
-		l.emit(itemPOW)
+		l.emit(POW)
 	case r == '=':
 		if t := l.peek(); t == '=' {
 			l.next()
-			l.emit(itemEQL)
+			l.emit(EQL)
 		} else if t == '~' {
 			return l.errorf("unexpected character after '=': %q", t)
 		} else {
-			l.emit(itemAssign)
+			l.emit(ASSIGN)
 		}
 	case r == '!':
 		if t := l.next(); t == '=' {
-			l.emit(itemNEQ)
+			l.emit(NEQ)
 		} else {
 			return l.errorf("unexpected character after '!': %q", t)
 		}
 	case r == '<':
 		if t := l.peek(); t == '=' {
 			l.next()
-			l.emit(itemLTE)
+			l.emit(LTE)
 		} else {
-			l.emit(itemLSS)
+			l.emit(LSS)
 		}
 	case r == '>':
 		if t := l.peek(); t == '=' {
 			l.next()
-			l.emit(itemGTE)
+			l.emit(GTE)
 		} else {
-			l.emit(itemGTR)
+			l.emit(GTR)
 		}
 	case isDigit(r) || (r == '.' && isDigit(l.peek())):
 		l.backup()
@@ -527,21 +461,21 @@ func lexStatements(l *lexer) stateFn {
 		if l.gotColon {
 			return l.errorf("unexpected colon %q", r)
 		}
-		l.emit(itemColon)
+		l.emit(COLON)
 		l.gotColon = true
 	case r == '(':
-		l.emit(itemLeftParen)
+		l.emit(LEFT_PAREN)
 		l.parenDepth++
 		return lexStatements
 	case r == ')':
-		l.emit(itemRightParen)
+		l.emit(RIGHT_PAREN)
 		l.parenDepth--
 		if l.parenDepth < 0 {
 			return l.errorf("unexpected right parenthesis %q", r)
 		}
 		return lexStatements
 	case r == '{':
-		l.emit(itemLeftBrace)
+		l.emit(LEFT_BRACE)
 		l.braceOpen = true
 		return lexInsideBraces(l)
 	case r == '[':
@@ -549,14 +483,17 @@ func lexStatements(l *lexer) stateFn {
 			return l.errorf("unexpected left bracket %q", r)
 		}
 		l.gotColon = false
-		l.emit(itemLeftBracket)
+		l.emit(LEFT_BRACKET)
+		if isSpace(l.peek()) {
+			skipSpaces(l)
+		}
 		l.bracketOpen = true
 		return lexDuration
 	case r == ']':
 		if !l.bracketOpen {
 			return l.errorf("unexpected right bracket %q", r)
 		}
-		l.emit(itemRightBracket)
+		l.emit(RIGHT_BRACKET)
 		l.bracketOpen = false
 
 	default:
@@ -567,7 +504,7 @@ func lexStatements(l *lexer) stateFn {
 
 // lexInsideBraces scans the inside of a vector selector. Keywords are ignored and
 // scanned as identifiers.
-func lexInsideBraces(l *lexer) stateFn {
+func lexInsideBraces(l *Lexer) stateFn {
 	if strings.HasPrefix(l.input[l.pos:], lineComment) {
 		return lexLineComment
 	}
@@ -581,7 +518,7 @@ func lexInsideBraces(l *lexer) stateFn {
 		l.backup()
 		return lexIdentifier
 	case r == ',':
-		l.emit(itemComma)
+		l.emit(COMMA)
 	case r == '"' || r == '\'':
 		l.stringOpen = r
 		return lexString
@@ -590,24 +527,24 @@ func lexInsideBraces(l *lexer) stateFn {
 		return lexRawString
 	case r == '=':
 		if l.next() == '~' {
-			l.emit(itemEQLRegex)
+			l.emit(EQL_REGEX)
 			break
 		}
 		l.backup()
-		l.emit(itemEQL)
+		l.emit(EQL)
 	case r == '!':
 		switch nr := l.next(); {
 		case nr == '~':
-			l.emit(itemNEQRegex)
+			l.emit(NEQ_REGEX)
 		case nr == '=':
-			l.emit(itemNEQ)
+			l.emit(NEQ)
 		default:
 			return l.errorf("unexpected character after '!' inside braces: %q", nr)
 		}
 	case r == '{':
 		return l.errorf("unexpected left brace %q", r)
 	case r == '}':
-		l.emit(itemRightBrace)
+		l.emit(RIGHT_BRACE)
 		l.braceOpen = false
 
 		if l.seriesDesc {
@@ -621,27 +558,27 @@ func lexInsideBraces(l *lexer) stateFn {
 }
 
 // lexValueSequence scans a value sequence of a series description.
-func lexValueSequence(l *lexer) stateFn {
+func lexValueSequence(l *Lexer) stateFn {
 	switch r := l.next(); {
 	case r == eof:
 		return lexStatements
 	case isSpace(r):
-		l.emit(itemSpace)
+		l.emit(SPACE)
 		lexSpace(l)
 	case r == '+':
-		l.emit(itemADD)
+		l.emit(ADD)
 	case r == '-':
-		l.emit(itemSUB)
+		l.emit(SUB)
 	case r == 'x':
-		l.emit(itemTimes)
+		l.emit(TIMES)
 	case r == '_':
-		l.emit(itemBlank)
+		l.emit(BLANK)
 	case isDigit(r) || (r == '.' && isDigit(l.peek())):
 		l.backup()
 		lexNumber(l)
 	case isAlpha(r):
 		l.backup()
-		// We might lex invalid items here but this will be caught by the parser.
+		// We might lex invalid Items here but this will be caught by the parser.
 		return lexKeywordOrIdentifier
 	default:
 		return l.errorf("unexpected character in series sequence: %q", r)
@@ -657,7 +594,7 @@ func lexValueSequence(l *lexer) stateFn {
 // package of the Go standard library to work for Prometheus-style strings.
 // None of the actual escaping/quoting logic was changed in this function - it
 // was only modified to integrate with our lexer.
-func lexEscape(l *lexer) {
+func lexEscape(l *Lexer) {
 	var n int
 	var base, max uint32
 
@@ -715,8 +652,16 @@ func digitVal(ch rune) int {
 	return 16 // Larger than any legal digit val.
 }
 
+// skipSpaces skips the spaces until a non-space is encountered.
+func skipSpaces(l *Lexer) {
+	for isSpace(l.peek()) {
+		l.next()
+	}
+	l.ignore()
+}
+
 // lexString scans a quoted string. The initial quote has already been seen.
-func lexString(l *lexer) stateFn {
+func lexString(l *Lexer) stateFn {
 Loop:
 	for {
 		switch l.next() {
@@ -730,12 +675,12 @@ Loop:
 			break Loop
 		}
 	}
-	l.emit(itemString)
+	l.emit(STRING)
 	return lexStatements
 }
 
 // lexRawString scans a raw quoted string. The initial quote has already been seen.
-func lexRawString(l *lexer) stateFn {
+func lexRawString(l *Lexer) stateFn {
 Loop:
 	for {
 		switch l.next() {
@@ -747,12 +692,12 @@ Loop:
 			break Loop
 		}
 	}
-	l.emit(itemString)
+	l.emit(STRING)
 	return lexStatements
 }
 
 // lexSpace scans a run of space characters. One space has already been seen.
-func lexSpace(l *lexer) stateFn {
+func lexSpace(l *Lexer) stateFn {
 	for isSpace(l.peek()) {
 		l.next()
 	}
@@ -761,17 +706,17 @@ func lexSpace(l *lexer) stateFn {
 }
 
 // lexLineComment scans a line comment. Left comment marker is known to be present.
-func lexLineComment(l *lexer) stateFn {
+func lexLineComment(l *Lexer) stateFn {
 	l.pos += Pos(len(lineComment))
 	for r := l.next(); !isEndOfLine(r) && r != eof; {
 		r = l.next()
 	}
 	l.backup()
-	l.emit(itemComment)
+	l.emit(COMMENT)
 	return lexStatements
 }
 
-func lexDuration(l *lexer) stateFn {
+func lexDuration(l *Lexer) stateFn {
 	if l.scanNumber() {
 		return l.errorf("missing unit character in duration")
 	}
@@ -781,25 +726,25 @@ func lexDuration(l *lexer) stateFn {
 			return l.errorf("bad duration syntax: %q", l.input[l.start:l.pos])
 		}
 		l.backup()
-		l.emit(itemDuration)
+		l.emit(DURATION)
 		return lexStatements
 	}
 	return l.errorf("bad duration syntax: %q", l.input[l.start:l.pos])
 }
 
 // lexNumber scans a number: decimal, hex, oct or float.
-func lexNumber(l *lexer) stateFn {
+func lexNumber(l *Lexer) stateFn {
 	if !l.scanNumber() {
 		return l.errorf("bad number syntax: %q", l.input[l.start:l.pos])
 	}
-	l.emit(itemNumber)
+	l.emit(NUMBER)
 	return lexStatements
 }
 
-// lexNumberOrDuration scans a number or a duration item.
-func lexNumberOrDuration(l *lexer) stateFn {
+// lexNumberOrDuration scans a number or a duration Item.
+func lexNumberOrDuration(l *Lexer) stateFn {
 	if l.scanNumber() {
-		l.emit(itemNumber)
+		l.emit(NUMBER)
 		return lexStatements
 	}
 	// Next two chars must be a valid unit and a non-alphanumeric.
@@ -808,15 +753,15 @@ func lexNumberOrDuration(l *lexer) stateFn {
 			return l.errorf("bad number or duration syntax: %q", l.input[l.start:l.pos])
 		}
 		l.backup()
-		l.emit(itemDuration)
+		l.emit(DURATION)
 		return lexStatements
 	}
 	return l.errorf("bad number or duration syntax: %q", l.input[l.start:l.pos])
 }
 
-// scanNumber scans numbers of different formats. The scanned item is
+// scanNumber scans numbers of different formats. The scanned Item is
 // not necessarily a valid number. This case is caught by the parser.
-func (l *lexer) scanNumber() bool {
+func (l *Lexer) scanNumber() bool {
 	digits := "0123456789"
 	// Disallow hexadecimal in series descriptions as the syntax is ambiguous.
 	if !l.seriesDesc && l.accept("0") && l.accept("xX") {
@@ -840,19 +785,19 @@ func (l *lexer) scanNumber() bool {
 
 // lexIdentifier scans an alphanumeric identifier. The next character
 // is known to be a letter.
-func lexIdentifier(l *lexer) stateFn {
+func lexIdentifier(l *Lexer) stateFn {
 	for isAlphaNumeric(l.next()) {
 		// absorb
 	}
 	l.backup()
-	l.emit(itemIdentifier)
+	l.emit(IDENTIFIER)
 	return lexStatements
 }
 
 // lexKeywordOrIdentifier scans an alphanumeric identifier which may contain
-// a colon rune. If the identifier is a keyword the respective keyword item
+// a colon rune. If the identifier is a keyword the respective keyword Item
 // is scanned.
-func lexKeywordOrIdentifier(l *lexer) stateFn {
+func lexKeywordOrIdentifier(l *Lexer) stateFn {
 Loop:
 	for {
 		switch r := l.next(); {
@@ -864,9 +809,9 @@ Loop:
 			if kw, ok := key[strings.ToLower(word)]; ok {
 				l.emit(kw)
 			} else if !strings.Contains(word, ":") {
-				l.emit(itemIdentifier)
+				l.emit(IDENTIFIER)
 			} else {
-				l.emit(itemMetricIdentifier)
+				l.emit(METRIC_IDENTIFIER)
 			}
 			break Loop
 		}
