@@ -23,17 +23,25 @@ import (
 	"strings"
 	"time"
 
+	"github.com/pkg/errors"
 	"github.com/prometheus/common/model"
+
 	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/pkg/value"
-
 	"github.com/prometheus/prometheus/util/strutil"
 )
 
 type parser struct {
-	lex       *lexer
-	token     [3]item
-	peekCount int
+	lex     *Lexer
+	token   Item
+	peeking bool
+
+	inject    Item
+	injecting bool
+
+	switchSymbols []ItemType
+
+	generatedParserResult interface{}
 }
 
 // ParseErr wraps a parsing error with line and position context.
@@ -45,10 +53,7 @@ type ParseErr struct {
 }
 
 func (e *ParseErr) Error() string {
-	if e.Line == 0 {
-		return fmt.Sprintf("parse error at char %d: %s", e.Pos, e.Err)
-	}
-	return fmt.Sprintf("parse error at line %d, char %d: %s", e.Line, e.Pos, e.Err)
+	return fmt.Sprintf("%d:%d: parse error: %s", e.Line+1, e.Pos, e.Err)
 }
 
 // ParseExpr returns the expression parsed from the input.
@@ -68,11 +73,7 @@ func ParseMetric(input string) (m labels.Labels, err error) {
 	p := newParser(input)
 	defer p.recover(&err)
 
-	m = p.metric()
-	if p.peek().typ != itemEOF {
-		p.errorf("could not parse remaining input %.15q...", p.lex.input[p.lex.lastPos:])
-	}
-	return m, nil
+	return p.parseGenerated(START_METRIC, []ItemType{RIGHT_BRACE, EOF}).(labels.Labels), nil
 }
 
 // ParseMetricSelector parses the provided textual metric selector into a list of
@@ -82,11 +83,11 @@ func ParseMetricSelector(input string) (m []*labels.Matcher, err error) {
 	defer p.recover(&err)
 
 	name := ""
-	if t := p.peek().typ; t == itemMetricIdentifier || t == itemIdentifier {
-		name = p.next().val
+	if t := p.peek().Typ; t == METRIC_IDENTIFIER || t == IDENTIFIER {
+		name = p.next().Val
 	}
 	vs := p.VectorSelector(name)
-	if p.peek().typ != itemEOF {
+	if p.peek().Typ != EOF {
 		p.errorf("could not parse remaining input %.15q...", p.lex.input[p.lex.lastPos:])
 	}
 	return vs.LabelMatchers, nil
@@ -95,7 +96,7 @@ func ParseMetricSelector(input string) (m []*labels.Matcher, err error) {
 // newParser returns a new parser.
 func newParser(input string) *parser {
 	p := &parser{
-		lex: lex(input),
+		lex: Lex(input),
 	}
 	return p
 }
@@ -104,10 +105,7 @@ func newParser(input string) *parser {
 func (p *parser) parseExpr() (expr Expr, err error) {
 	defer p.recover(&err)
 
-	for p.peek().typ != itemEOF {
-		if p.peek().typ == itemComment {
-			continue
-		}
+	for p.peek().Typ != EOF {
 		if expr != nil {
 			p.errorf("could not parse remaining input %.15q...", p.lex.input[p.lex.lastPos:])
 		}
@@ -149,20 +147,20 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 
 	const ctx = "series values"
 	for {
-		for p.peek().typ == itemSpace {
+		for p.peek().Typ == SPACE {
 			p.next()
 		}
-		if p.peek().typ == itemEOF {
+		if p.peek().Typ == EOF {
 			break
 		}
 
 		// Extract blanks.
-		if p.peek().typ == itemBlank {
+		if p.peek().Typ == BLANK {
 			p.next()
 			times := uint64(1)
-			if p.peek().typ == itemTimes {
+			if p.peek().Typ == TIMES {
 				p.next()
-				times, err = strconv.ParseUint(p.expect(itemNumber, ctx).val, 10, 64)
+				times, err = strconv.ParseUint(p.expect(NUMBER, ctx).Val, 10, 64)
 				if err != nil {
 					p.errorf("invalid repetition in %s: %s", ctx, err)
 				}
@@ -172,7 +170,7 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 			}
 			// This is to ensure that there is a space between this and the next number.
 			// This is especially required if the next number is negative.
-			if t := p.expectOneOf(itemSpace, itemEOF, ctx).typ; t == itemEOF {
+			if t := p.expectOneOf(SPACE, EOF, ctx).Typ; t == EOF {
 				break
 			}
 			continue
@@ -180,15 +178,15 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 
 		// Extract values.
 		sign := 1.0
-		if t := p.peek().typ; t == itemSUB || t == itemADD {
-			if p.next().typ == itemSUB {
+		if t := p.peek().Typ; t == SUB || t == ADD {
+			if p.next().Typ == SUB {
 				sign = -1
 			}
 		}
 		var k float64
-		if t := p.peek().typ; t == itemNumber {
-			k = sign * p.number(p.expect(itemNumber, ctx).val)
-		} else if t == itemIdentifier && p.peek().val == "stale" {
+		if t := p.peek().Typ; t == NUMBER {
+			k = sign * p.number(p.expect(NUMBER, ctx).Val)
+		} else if t == IDENTIFIER && p.peek().Val == "stale" {
 			p.next()
 			k = math.Float64frombits(value.StaleNaN)
 		} else {
@@ -199,24 +197,24 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 		})
 
 		// If there are no offset repetitions specified, proceed with the next value.
-		if t := p.peek(); t.typ == itemSpace {
+		if t := p.peek(); t.Typ == SPACE {
 			// This ensures there is a space between every value.
 			continue
-		} else if t.typ == itemEOF {
+		} else if t.Typ == EOF {
 			break
-		} else if t.typ != itemADD && t.typ != itemSUB {
+		} else if t.Typ != ADD && t.Typ != SUB {
 			p.errorf("expected next value or relative expansion in %s but got %s (value: %s)", ctx, t.desc(), p.peek())
 		}
 
 		// Expand the repeated offsets into values.
 		sign = 1.0
-		if p.next().typ == itemSUB {
+		if p.next().Typ == SUB {
 			sign = -1.0
 		}
-		offset := sign * p.number(p.expect(itemNumber, ctx).val)
-		p.expect(itemTimes, ctx)
+		offset := sign * p.number(p.expect(NUMBER, ctx).Val)
+		p.expect(TIMES, ctx)
 
-		times, err := strconv.ParseUint(p.expect(itemNumber, ctx).val, 10, 64)
+		times, err := strconv.ParseUint(p.expect(NUMBER, ctx).Val, 10, 64)
 		if err != nil {
 			p.errorf("invalid repetition in %s: %s", ctx, err)
 		}
@@ -230,7 +228,7 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 		// This is to ensure that there is a space between this expanding notation
 		// and the next number. This is especially required if the next number
 		// is negative.
-		if t := p.expectOneOf(itemSpace, itemEOF, ctx).typ; t == itemEOF {
+		if t := p.expectOneOf(SPACE, EOF, ctx).Typ; t == EOF {
 			break
 		}
 	}
@@ -246,47 +244,48 @@ func (p *parser) typecheck(node Node) (err error) {
 }
 
 // next returns the next token.
-func (p *parser) next() item {
-	if p.peekCount > 0 {
-		p.peekCount--
-	} else {
-		t := p.lex.nextItem()
+func (p *parser) next() Item {
+	if !p.peeking {
+		t := p.lex.NextItem()
 		// Skip comments.
-		for t.typ == itemComment {
-			t = p.lex.nextItem()
+		for t.Typ == COMMENT {
+			t = p.lex.NextItem()
 		}
-		p.token[0] = t
+		p.token = t
 	}
-	if p.token[p.peekCount].typ == itemError {
-		p.errorf("%s", p.token[p.peekCount].val)
+
+	p.peeking = false
+
+	if p.token.Typ == ERROR {
+		p.errorf("%s", p.token.Val)
 	}
-	return p.token[p.peekCount]
+	return p.token
 }
 
 // peek returns but does not consume the next token.
-func (p *parser) peek() item {
-	if p.peekCount > 0 {
-		return p.token[p.peekCount-1]
+func (p *parser) peek() Item {
+	if p.peeking {
+		return p.token
 	}
-	p.peekCount = 1
+	p.peeking = true
 
-	t := p.lex.nextItem()
+	t := p.lex.NextItem()
 	// Skip comments.
-	for t.typ == itemComment {
-		t = p.lex.nextItem()
+	for t.Typ == COMMENT {
+		t = p.lex.NextItem()
 	}
-	p.token[0] = t
-	return p.token[0]
+	p.token = t
+	return p.token
 }
 
 // backup backs the input stream up one token.
 func (p *parser) backup() {
-	p.peekCount++
+	p.peeking = true
 }
 
 // errorf formats the error and terminates processing.
 func (p *parser) errorf(format string, args ...interface{}) {
-	p.error(fmt.Errorf(format, args...))
+	p.error(errors.Errorf(format, args...))
 }
 
 // error terminates processing.
@@ -303,43 +302,97 @@ func (p *parser) error(err error) {
 }
 
 // expect consumes the next token and guarantees it has the required type.
-func (p *parser) expect(exp ItemType, context string) item {
+func (p *parser) expect(exp ItemType, context string) Item {
 	token := p.next()
-	if token.typ != exp {
+	if token.Typ != exp {
 		p.errorf("unexpected %s in %s, expected %s", token.desc(), context, exp.desc())
 	}
 	return token
 }
 
 // expectOneOf consumes the next token and guarantees it has one of the required types.
-func (p *parser) expectOneOf(exp1, exp2 ItemType, context string) item {
+func (p *parser) expectOneOf(exp1, exp2 ItemType, context string) Item {
 	token := p.next()
-	if token.typ != exp1 && token.typ != exp2 {
+	if token.Typ != exp1 && token.Typ != exp2 {
 		p.errorf("unexpected %s in %s, expected %s or %s", token.desc(), context, exp1.desc(), exp2.desc())
 	}
 	return token
 }
 
-var errUnexpected = fmt.Errorf("unexpected error")
+var errUnexpected = errors.New("unexpected error")
 
 // recover is the handler that turns panics into returns from the top level of Parse.
 func (p *parser) recover(errp *error) {
 	e := recover()
-	if e != nil {
-		if _, ok := e.(runtime.Error); ok {
-			// Print the stack trace but do not inhibit the running application.
-			buf := make([]byte, 64<<10)
-			buf = buf[:runtime.Stack(buf, false)]
+	if _, ok := e.(runtime.Error); ok {
+		// Print the stack trace but do not inhibit the running application.
+		buf := make([]byte, 64<<10)
+		buf = buf[:runtime.Stack(buf, false)]
 
-			fmt.Fprintf(os.Stderr, "parser panic: %v\n%s", e, buf)
-			*errp = errUnexpected
-		} else {
-			*errp = e.(error)
-		}
+		fmt.Fprintf(os.Stderr, "parser panic: %v\n%s", e, buf)
+		*errp = errUnexpected
+	} else if e != nil {
+		*errp = e.(error)
 	}
 	p.lex.close()
 }
 
+// Lex is expected by the yyLexer interface of the yacc generated parser.
+// It writes the next Item provided by the lexer to the provided pointer address.
+// Comments are skipped.
+//
+// The yyLexer interface is currently implemented by the parser to allow
+// the generated and non-generated parts to work together with regards to lookahead
+// and error handling.
+//
+// For more information, see https://godoc.org/golang.org/x/tools/cmd/goyacc.
+func (p *parser) Lex(lval *yySymType) int {
+	if p.injecting {
+		lval.item = p.inject
+		p.injecting = false
+	} else {
+		lval.item = p.next()
+	}
+
+	typ := lval.item.Typ
+
+	for _, t := range p.switchSymbols {
+		if t == typ {
+			p.InjectItem(0)
+		}
+	}
+
+	return int(typ)
+}
+
+// Error is expected by the yyLexer interface of the yacc generated parser.
+//
+// It is a no-op since the parsers error routines are triggered
+// by mechanisms that allow more fine grained control
+// For more information, see https://godoc.org/golang.org/x/tools/cmd/goyacc.
+func (p *parser) Error(e string) {
+}
+
+// InjectItem allows injecting a single Item at the beginning of the token stream
+// consumed by the generated parser.
+// This allows having multiple start symbols as described in
+// https://www.gnu.org/software/bison/manual/html_node/Multiple-start_002dsymbols.html .
+// Only the Lex function used by the generated parser is affected by this injected Item.
+// Trying to inject when a previously injected Item has not yet been consumed will panic.
+// Only Item types that are supposed to be used as start symbols are allowed as an argument.
+func (p *parser) InjectItem(typ ItemType) {
+	if p.injecting {
+		panic("cannot inject multiple Items into the token stream")
+	}
+
+	if typ != 0 && (typ <= startSymbolsStart || typ >= startSymbolsEnd) {
+		panic("cannot inject symbol that isn't start symbol")
+	}
+
+	p.inject = Item{Typ: typ}
+	p.injecting = true
+}
+
 // expr parses any expression.
 func (p *parser) expr() Expr {
 	// Parse the starting expression.
@@ -349,14 +402,14 @@ func (p *parser) expr() Expr {
 	// on the operators' precedence.
 	for {
 		// If the next token is not an operator the expression is done.
-		op := p.peek().typ
+		op := p.peek().Typ
 		if !op.isOperator() {
 			// Check for subquery.
-			if op == itemLeftBracket {
+			if op == LEFT_BRACKET {
 				expr = p.subqueryOrRangeSelector(expr, false)
 				if s, ok := expr.(*SubqueryExpr); ok {
 					// Parse optional offset.
-					if p.peek().typ == itemOffset {
+					if p.peek().Typ == OFFSET {
 						offset := p.offset()
 						s.Offset = offset
 					}
@@ -377,7 +430,7 @@ func (p *parser) expr() Expr {
 
 		returnBool := false
 		// Parse bool modifier.
-		if p.peek().typ == itemBool {
+		if p.peek().Typ == BOOL {
 			if !op.isComparisonOperator() {
 				p.errorf("bool modifier can only be used on comparison operators")
 			}
@@ -386,22 +439,22 @@ func (p *parser) expr() Expr {
 		}
 
 		// Parse ON/IGNORING clause.
-		if p.peek().typ == itemOn || p.peek().typ == itemIgnoring {
-			if p.peek().typ == itemOn {
+		if p.peek().Typ == ON || p.peek().Typ == IGNORING {
+			if p.peek().Typ == ON {
 				vecMatching.On = true
 			}
 			p.next()
 			vecMatching.MatchingLabels = p.labels()
 
 			// Parse grouping.
-			if t := p.peek().typ; t == itemGroupLeft || t == itemGroupRight {
+			if t := p.peek().Typ; t == GROUP_LEFT || t == GROUP_RIGHT {
 				p.next()
-				if t == itemGroupLeft {
+				if t == GROUP_LEFT {
 					vecMatching.Card = CardManyToOne
 				} else {
 					vecMatching.Card = CardOneToMany
 				}
-				if p.peek().typ == itemLeftParen {
+				if p.peek().Typ == LEFT_PAREN {
 					vecMatching.Include = p.labels()
 				}
 			}
@@ -457,36 +510,36 @@ func (p *parser) balance(lhs Expr, op ItemType, rhs Expr, vecMatching *VectorMat
 //		<Vector_selector> | <Matrix_selector> | (+|-) <number_literal> | '(' <expr> ')'
 //
 func (p *parser) unaryExpr() Expr {
-	switch t := p.peek(); t.typ {
-	case itemADD, itemSUB:
+	switch t := p.peek(); t.Typ {
+	case ADD, SUB:
 		p.next()
 		e := p.unaryExpr()
 
 		// Simplify unary expressions for number literals.
 		if nl, ok := e.(*NumberLiteral); ok {
-			if t.typ == itemSUB {
+			if t.Typ == SUB {
 				nl.Val *= -1
 			}
 			return nl
 		}
-		return &UnaryExpr{Op: t.typ, Expr: e}
+		return &UnaryExpr{Op: t.Typ, Expr: e}
 
-	case itemLeftParen:
+	case LEFT_PAREN:
 		p.next()
 		e := p.expr()
-		p.expect(itemRightParen, "paren expression")
+		p.expect(RIGHT_PAREN, "paren expression")
 
 		return &ParenExpr{Expr: e}
 	}
 	e := p.primaryExpr()
 
 	// Expression might be followed by a range selector.
-	if p.peek().typ == itemLeftBracket {
+	if p.peek().Typ == LEFT_BRACKET {
 		e = p.subqueryOrRangeSelector(e, true)
 	}
 
 	// Parse optional offset.
-	if p.peek().typ == itemOffset {
+	if p.peek().Typ == OFFSET {
 		offset := p.offset()
 
 		switch s := e.(type) {
@@ -520,16 +573,16 @@ func (p *parser) subqueryOrRangeSelector(expr Expr, checkRange bool) Expr {
 	var erange time.Duration
 	var err error
 
-	erangeStr := p.expect(itemDuration, ctx).val
+	erangeStr := p.expect(DURATION, ctx).Val
 	erange, err = parseDuration(erangeStr)
 	if err != nil {
 		p.error(err)
 	}
 
-	var itm item
+	var itm Item
 	if checkRange {
-		itm = p.expectOneOf(itemRightBracket, itemColon, ctx)
-		if itm.typ == itemRightBracket {
+		itm = p.expectOneOf(RIGHT_BRACKET, COLON, ctx)
+		if itm.Typ == RIGHT_BRACKET {
 			// Range selector.
 			vs, ok := expr.(*VectorSelector)
 			if !ok {
@@ -542,20 +595,20 @@ func (p *parser) subqueryOrRangeSelector(expr Expr, checkRange bool) Expr {
 			}
 		}
 	} else {
-		itm = p.expect(itemColon, ctx)
+		itm = p.expect(COLON, ctx)
 	}
 
 	// Subquery.
 	var estep time.Duration
 
-	itm = p.expectOneOf(itemRightBracket, itemDuration, ctx)
-	if itm.typ == itemDuration {
-		estepStr := itm.val
+	itm = p.expectOneOf(RIGHT_BRACKET, DURATION, ctx)
+	if itm.Typ == DURATION {
+		estepStr := itm.Val
 		estep, err = parseDuration(estepStr)
 		if err != nil {
 			p.error(err)
 		}
-		p.expect(itemRightBracket, ctx)
+		p.expect(RIGHT_BRACKET, ctx)
 	}
 
 	return &SubqueryExpr{
@@ -584,29 +637,29 @@ func (p *parser) number(val string) float64 {
 //
 func (p *parser) primaryExpr() Expr {
 	switch t := p.next(); {
-	case t.typ == itemNumber:
-		f := p.number(t.val)
+	case t.Typ == NUMBER:
+		f := p.number(t.Val)
 		return &NumberLiteral{f}
 
-	case t.typ == itemString:
-		return &StringLiteral{p.unquoteString(t.val)}
+	case t.Typ == STRING:
+		return &StringLiteral{p.unquoteString(t.Val)}
 
-	case t.typ == itemLeftBrace:
+	case t.Typ == LEFT_BRACE:
 		// Metric selector without metric name.
 		p.backup()
 		return p.VectorSelector("")
 
-	case t.typ == itemIdentifier:
+	case t.Typ == IDENTIFIER:
 		// Check for function call.
-		if p.peek().typ == itemLeftParen {
-			return p.call(t.val)
+		if p.peek().Typ == LEFT_PAREN {
+			return p.call(t.Val)
 		}
 		fallthrough // Else metric selector.
 
-	case t.typ == itemMetricIdentifier:
-		return p.VectorSelector(t.val)
+	case t.Typ == METRIC_IDENTIFIER:
+		return p.VectorSelector(t.Val)
 
-	case t.typ.isAggregator():
+	case t.Typ.isAggregator():
 		p.backup()
 		return p.aggrExpr()
 
@@ -623,24 +676,24 @@ func (p *parser) primaryExpr() Expr {
 func (p *parser) labels() []string {
 	const ctx = "grouping opts"
 
-	p.expect(itemLeftParen, ctx)
+	p.expect(LEFT_PAREN, ctx)
 
 	labels := []string{}
-	if p.peek().typ != itemRightParen {
+	if p.peek().Typ != RIGHT_PAREN {
 		for {
 			id := p.next()
-			if !isLabel(id.val) {
+			if !isLabel(id.Val) {
 				p.errorf("unexpected %s in %s, expected label", id.desc(), ctx)
 			}
-			labels = append(labels, id.val)
+			labels = append(labels, id.Val)
 
-			if p.peek().typ != itemComma {
+			if p.peek().Typ != COMMA {
 				break
 			}
 			p.next()
 		}
 	}
-	p.expect(itemRightParen, ctx)
+	p.expect(RIGHT_PAREN, ctx)
 
 	return labels
 }
@@ -654,7 +707,7 @@ func (p *parser) aggrExpr() *AggregateExpr {
 	const ctx = "aggregation"
 
 	agop := p.next()
-	if !agop.typ.isAggregator() {
+	if !agop.Typ.isAggregator() {
 		p.errorf("expected aggregation operator but got %s", agop)
 	}
 	var grouping []string
@@ -662,8 +715,8 @@ func (p *parser) aggrExpr() *AggregateExpr {
 
 	modifiersFirst := false
 
-	if t := p.peek().typ; t == itemBy || t == itemWithout {
-		if t == itemWithout {
+	if t := p.peek().Typ; t == BY || t == WITHOUT {
+		if t == WITHOUT {
 			without = true
 		}
 		p.next()
@@ -671,21 +724,21 @@ func (p *parser) aggrExpr() *AggregateExpr {
 		modifiersFirst = true
 	}
 
-	p.expect(itemLeftParen, ctx)
+	p.expect(LEFT_PAREN, ctx)
 	var param Expr
-	if agop.typ.isAggregatorWithParam() {
+	if agop.Typ.isAggregatorWithParam() {
 		param = p.expr()
-		p.expect(itemComma, ctx)
+		p.expect(COMMA, ctx)
 	}
 	e := p.expr()
-	p.expect(itemRightParen, ctx)
+	p.expect(RIGHT_PAREN, ctx)
 
 	if !modifiersFirst {
-		if t := p.peek().typ; t == itemBy || t == itemWithout {
+		if t := p.peek().Typ; t == BY || t == WITHOUT {
 			if len(grouping) > 0 {
 				p.errorf("aggregation must only contain one grouping clause")
 			}
-			if t == itemWithout {
+			if t == WITHOUT {
 				without = true
 			}
 			p.next()
@@ -694,7 +747,7 @@ func (p *parser) aggrExpr() *AggregateExpr {
 	}
 
 	return &AggregateExpr{
-		Op:       agop.typ,
+		Op:       agop.Typ,
 		Expr:     e,
 		Param:    param,
 		Grouping: grouping,
@@ -714,9 +767,9 @@ func (p *parser) call(name string) *Call {
 		p.errorf("unknown function with name %q", name)
 	}
 
-	p.expect(itemLeftParen, ctx)
+	p.expect(LEFT_PAREN, ctx)
 	// Might be call without args.
-	if p.peek().typ == itemRightParen {
+	if p.peek().Typ == RIGHT_PAREN {
 		p.next() // Consume.
 		return &Call{fn, nil}
 	}
@@ -727,14 +780,14 @@ func (p *parser) call(name string) *Call {
 		args = append(args, e)
 
 		// Terminate if no more arguments.
-		if p.peek().typ != itemComma {
+		if p.peek().Typ != COMMA {
 			break
 		}
 		p.next()
 	}
 
 	// Call must be closed.
-	p.expect(itemRightParen, ctx)
+	p.expect(RIGHT_PAREN, ctx)
 
 	return &Call{Func: fn, Args: args}
 }
@@ -744,90 +797,7 @@ func (p *parser) call(name string) *Call {
 //		'{' [ <labelname> '=' <match_string>, ... ] '}'
 //
 func (p *parser) labelSet() labels.Labels {
-	set := []labels.Label{}
-	for _, lm := range p.labelMatchers(itemEQL) {
-		set = append(set, labels.Label{Name: lm.Name, Value: lm.Value})
-	}
-	return labels.New(set...)
-}
-
-// labelMatchers parses a set of label matchers.
-//
-//		'{' [ <labelname> <match_op> <match_string>, ... ] '}'
-//
-func (p *parser) labelMatchers(operators ...ItemType) []*labels.Matcher {
-	const ctx = "label matching"
-
-	matchers := []*labels.Matcher{}
-
-	p.expect(itemLeftBrace, ctx)
-
-	// Check if no matchers are provided.
-	if p.peek().typ == itemRightBrace {
-		p.next()
-		return matchers
-	}
-
-	for {
-		label := p.expect(itemIdentifier, ctx)
-
-		op := p.next().typ
-		if !op.isOperator() {
-			p.errorf("expected label matching operator but got %s", op)
-		}
-		var validOp = false
-		for _, allowedOp := range operators {
-			if op == allowedOp {
-				validOp = true
-			}
-		}
-		if !validOp {
-			p.errorf("operator must be one of %q, is %q", operators, op)
-		}
-
-		val := p.unquoteString(p.expect(itemString, ctx).val)
-
-		// Map the item to the respective match type.
-		var matchType labels.MatchType
-		switch op {
-		case itemEQL:
-			matchType = labels.MatchEqual
-		case itemNEQ:
-			matchType = labels.MatchNotEqual
-		case itemEQLRegex:
-			matchType = labels.MatchRegexp
-		case itemNEQRegex:
-			matchType = labels.MatchNotRegexp
-		default:
-			p.errorf("item %q is not a metric match type", op)
-		}
-
-		m, err := labels.NewMatcher(matchType, label.val, val)
-		if err != nil {
-			p.error(err)
-		}
-
-		matchers = append(matchers, m)
-
-		if p.peek().typ == itemIdentifier {
-			p.errorf("missing comma before next identifier %q", p.peek().val)
-		}
-
-		// Terminate list if last matcher.
-		if p.peek().typ != itemComma {
-			break
-		}
-		p.next()
-
-		// Allow comma after each item in a multi-line listing.
-		if p.peek().typ == itemRightBrace {
-			break
-		}
-	}
-
-	p.expect(itemRightBrace, ctx)
-
-	return matchers
+	return p.parseGenerated(START_LABEL_SET, []ItemType{RIGHT_BRACE, EOF}).(labels.Labels)
 }
 
 // metric parses a metric.
@@ -839,15 +809,15 @@ func (p *parser) metric() labels.Labels {
 	name := ""
 	var m labels.Labels
 
-	t := p.peek().typ
-	if t == itemIdentifier || t == itemMetricIdentifier {
-		name = p.next().val
-		t = p.peek().typ
+	t := p.peek().Typ
+	if t == IDENTIFIER || t == METRIC_IDENTIFIER {
+		name = p.next().Val
+		t = p.peek().Typ
 	}
-	if t != itemLeftBrace && name == "" {
+	if t != LEFT_BRACE && name == "" {
 		p.errorf("missing metric name or metric selector")
 	}
-	if t == itemLeftBrace {
+	if t == LEFT_BRACE {
 		m = p.labelSet()
 	}
 	if name != "" {
@@ -865,9 +835,9 @@ func (p *parser) offset() time.Duration {
 	const ctx = "offset"
 
 	p.next()
-	offi := p.expect(itemDuration, ctx)
+	offi := p.expect(DURATION, ctx)
 
-	offset, err := parseDuration(offi.val)
+	offset, err := parseDuration(offi.Val)
 	if err != nil {
 		p.error(err)
 	}
@@ -881,14 +851,18 @@ func (p *parser) offset() time.Duration {
 //		[<metric_identifier>] <label_matchers>
 //
 func (p *parser) VectorSelector(name string) *VectorSelector {
-	var matchers []*labels.Matcher
+	ret := &VectorSelector{
+		Name: name,
+	}
 	// Parse label matching if any.
-	if t := p.peek(); t.typ == itemLeftBrace {
-		matchers = p.labelMatchers(itemEQL, itemNEQ, itemEQLRegex, itemNEQRegex)
+	if t := p.peek(); t.Typ == LEFT_BRACE {
+		p.generatedParserResult = ret
+
+		p.parseGenerated(START_LABELS, []ItemType{RIGHT_BRACE, EOF})
 	}
 	// Metric name must not be set in the label matchers and before at the same time.
 	if name != "" {
-		for _, m := range matchers {
+		for _, m := range ret.LabelMatchers {
 			if m.Name == labels.MetricName {
 				p.errorf("metric name must not be set twice: %q or %q", name, m.Value)
 			}
@@ -898,16 +872,16 @@ func (p *parser) VectorSelector(name string) *VectorSelector {
 		if err != nil {
 			panic(err) // Must not happen with metric.Equal.
 		}
-		matchers = append(matchers, m)
+		ret.LabelMatchers = append(ret.LabelMatchers, m)
 	}
 
-	if len(matchers) == 0 {
+	if len(ret.LabelMatchers) == 0 {
 		p.errorf("vector selector must contain label matchers or metric name")
 	}
 	// A Vector selector must contain at least one non-empty matcher to prevent
 	// implicit selection of all metrics (e.g. by a typo).
 	notEmpty := false
-	for _, lm := range matchers {
+	for _, lm := range ret.LabelMatchers {
 		if !lm.Matches("") {
 			notEmpty = true
 			break
@@ -917,10 +891,7 @@ func (p *parser) VectorSelector(name string) *VectorSelector {
 		p.errorf("vector selector must contain at least one non-empty matcher")
 	}
 
-	return &VectorSelector{
-		Name:          name,
-		LabelMatchers: matchers,
-	}
+	return ret
 }
 
 // expectType checks the type of the node and raises an error if it
@@ -935,7 +906,7 @@ func (p *parser) expectType(node Node, want ValueType, context string) {
 // check the types of the children of each node and raise an error
 // if they do not form a valid node.
 //
-// Some of these checks are redundant as the the parsing stage does not allow
+// Some of these checks are redundant as the parsing stage does not allow
 // them, but the costs are small and might reveal errors when making changes.
 func (p *parser) checkType(node Node) (typ ValueType) {
 	// For expressions the type is determined by their Type function.
@@ -970,10 +941,10 @@ func (p *parser) checkType(node Node) (typ ValueType) {
 			p.errorf("aggregation operator expected in aggregation expression but got %q", n.Op)
 		}
 		p.expectType(n.Expr, ValueTypeVector, "aggregation expression")
-		if n.Op == itemTopK || n.Op == itemBottomK || n.Op == itemQuantile {
+		if n.Op == TOPK || n.Op == BOTTOMK || n.Op == QUANTILE {
 			p.expectType(n.Param, ValueTypeScalar, "aggregation parameter")
 		}
-		if n.Op == itemCountValues {
+		if n.Op == COUNT_VALUES {
 			p.expectType(n.Param, ValueTypeString, "aggregation parameter")
 		}
 
@@ -1035,7 +1006,7 @@ func (p *parser) checkType(node Node) (typ ValueType) {
 		p.checkType(n.Expr)
 
 	case *UnaryExpr:
-		if n.Op != itemADD && n.Op != itemSUB {
+		if n.Op != ADD && n.Op != SUB {
 			p.errorf("only + and - operators allowed for unary expressions")
 		}
 		if t := p.checkType(n.Expr); t != ValueTypeScalar && t != ValueTypeVector {
@@ -1071,7 +1042,54 @@ func parseDuration(ds string) (time.Duration, error) {
 		return 0, err
 	}
 	if dur == 0 {
-		return 0, fmt.Errorf("duration must be greater than 0")
+		return 0, errors.New("duration must be greater than 0")
 	}
 	return time.Duration(dur), nil
 }
+
+// parseGenerated invokes the yacc generated parser.
+// The generated parser gets the provided startSymbol injected into
+// the lexer stream, based on which grammar will be used.
+//
+// The generated parser will consume the lexer Stream until one of the
+// tokens listed in switchSymbols is encountered. switchSymbols
+// should at least contain EOF
+func (p *parser) parseGenerated(startSymbol ItemType, switchSymbols []ItemType) interface{} {
+	p.InjectItem(startSymbol)
+
+	p.switchSymbols = switchSymbols
+
+	yyParse(p)
+
+	return p.generatedParserResult
+
+}
+
+func (p *parser) newLabelMatcher(label Item, operator Item, value Item) *labels.Matcher {
+	op := operator.Typ
+	val := p.unquoteString(value.Val)
+
+	// Map the Item to the respective match type.
+	var matchType labels.MatchType
+	switch op {
+	case EQL:
+		matchType = labels.MatchEqual
+	case NEQ:
+		matchType = labels.MatchNotEqual
+	case EQL_REGEX:
+		matchType = labels.MatchRegexp
+	case NEQ_REGEX:
+		matchType = labels.MatchNotRegexp
+	default:
+		// This should never happen, since the error should have been chaught
+		// by the generated parser.
+		panic("invalid operator")
+	}
+
+	m, err := labels.NewMatcher(matchType, label.Val, val)
+	if err != nil {
+		p.error(err)
+	}
+
+	return m
+}
