@@ -23,34 +23,32 @@ import (
 	"net/http"
 	"net/http/httptest"
 	"net/url"
-	"reflect"
 	"strings"
 	"sync"
 	"testing"
 	"time"
 
-	"github.com/prometheus/prometheus/pkg/relabel"
-
-	"github.com/prometheus/common/model"
-	"github.com/stretchr/testify/require"
-
+	"github.com/pkg/errors"
 	dto "github.com/prometheus/client_model/go"
+	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
 	"github.com/prometheus/prometheus/pkg/labels"
+	"github.com/prometheus/prometheus/pkg/relabel"
 	"github.com/prometheus/prometheus/pkg/textparse"
 	"github.com/prometheus/prometheus/pkg/timestamp"
 	"github.com/prometheus/prometheus/pkg/value"
 	"github.com/prometheus/prometheus/storage"
+	"github.com/prometheus/prometheus/util/teststorage"
 	"github.com/prometheus/prometheus/util/testutil"
 )
 
 func TestNewScrapePool(t *testing.T) {
 	var (
-		app = &nopAppendable{}
-		cfg = &config.ScrapeConfig{}
-		sp  = newScrapePool(cfg, app, nil)
+		app   = &nopAppendable{}
+		cfg   = &config.ScrapeConfig{}
+		sp, _ = newScrapePool(cfg, app, 0, nil)
 	)
 
 	if a, ok := sp.appendable.(*nopAppendable); !ok || a != app {
@@ -85,8 +83,8 @@ func TestDroppedTargetsList(t *testing.T) {
 				},
 			},
 		}
-		sp                     = newScrapePool(cfg, app, nil)
-		expectedLabelSetString = "{__address__=\"127.0.0.1:9090\", __metrics_path__=\"\", __scheme__=\"\", job=\"dropMe\"}"
+		sp, _                  = newScrapePool(cfg, app, 0, nil)
+		expectedLabelSetString = "{__address__=\"127.0.0.1:9090\", job=\"dropMe\"}"
 		expectedLength         = 1
 	)
 	sp.Sync(tgs)
@@ -151,6 +149,7 @@ func TestScrapePoolStop(t *testing.T) {
 		activeTargets: map[uint64]*Target{},
 		loops:         map[uint64]loop{},
 		cancel:        func() {},
+		client:        http.DefaultClient,
 	}
 	var mtx sync.Mutex
 	stopped := map[uint64]bool{}
@@ -196,17 +195,11 @@ func TestScrapePoolStop(t *testing.T) {
 	}
 
 	mtx.Lock()
-	if len(stopped) != numTargets {
-		t.Fatalf("Expected 20 stopped loops, got %d", len(stopped))
-	}
+	testutil.Equals(t, numTargets, len(stopped), "Unexpected number of stopped loops")
 	mtx.Unlock()
 
-	if len(sp.activeTargets) > 0 {
-		t.Fatalf("Targets were not cleared on stopping: %d left", len(sp.activeTargets))
-	}
-	if len(sp.loops) > 0 {
-		t.Fatalf("Loops were not cleared on stopping: %d left", len(sp.loops))
-	}
+	testutil.Assert(t, len(sp.activeTargets) == 0, "Targets were not cleared on stopping: %d left", len(sp.activeTargets))
+	testutil.Assert(t, len(sp.loops) == 0, "Loops were not cleared on stopping: %d left", len(sp.loops))
 }
 
 func TestScrapePoolReload(t *testing.T) {
@@ -221,19 +214,15 @@ func TestScrapePoolReload(t *testing.T) {
 	}
 	// On starting to run, new loops created on reload check whether their preceding
 	// equivalents have been stopped.
-	newLoop := func(_ *Target, s scraper, _ int, _ bool, _ []*relabel.Config) loop {
+	newLoop := func(opts scrapeLoopOptions) loop {
 		l := &testLoop{}
 		l.startFunc = func(interval, timeout time.Duration, errc chan<- error) {
-			if interval != 3*time.Second {
-				t.Errorf("Expected scrape interval %d but got %d", 3*time.Second, interval)
-			}
-			if timeout != 2*time.Second {
-				t.Errorf("Expected scrape timeout %d but got %d", 2*time.Second, timeout)
-			}
+			testutil.Equals(t, 3*time.Second, interval, "Unexpected scrape interval")
+			testutil.Equals(t, 2*time.Second, timeout, "Unexpected scrape timeout")
+
 			mtx.Lock()
-			if !stopped[s.(*targetScraper).hash()] {
-				t.Errorf("Scrape loop for %v not stopped yet", s.(*targetScraper))
-			}
+			targetScraper := opts.scraper.(*targetScraper)
+			testutil.Assert(t, stopped[targetScraper.hash()], "Scrape loop for %v not stopped yet", targetScraper)
 			mtx.Unlock()
 		}
 		return l
@@ -244,6 +233,7 @@ func TestScrapePoolReload(t *testing.T) {
 		loops:         map[uint64]loop{},
 		newLoop:       newLoop,
 		logger:        nil,
+		client:        http.DefaultClient,
 	}
 
 	// Reloading a scrape pool with a new scrape configuration must stop all scrape
@@ -291,57 +281,49 @@ func TestScrapePoolReload(t *testing.T) {
 	}
 
 	mtx.Lock()
-	if len(stopped) != numTargets {
-		t.Fatalf("Expected 20 stopped loops, got %d", len(stopped))
-	}
+	testutil.Equals(t, numTargets, len(stopped), "Unexpected number of stopped loops")
 	mtx.Unlock()
 
-	if !reflect.DeepEqual(sp.activeTargets, beforeTargets) {
-		t.Fatalf("Reloading affected target states unexpectedly")
-	}
-	if len(sp.loops) != numTargets {
-		t.Fatalf("Expected %d loops after reload but got %d", numTargets, len(sp.loops))
-	}
+	testutil.Equals(t, sp.activeTargets, beforeTargets, "Reloading affected target states unexpectedly")
+	testutil.Equals(t, numTargets, len(sp.loops), "Unexpected number of stopped loops after reload")
 }
 
 func TestScrapePoolAppender(t *testing.T) {
 	cfg := &config.ScrapeConfig{}
 	app := &nopAppendable{}
-	sp := newScrapePool(cfg, app, nil)
+	sp, _ := newScrapePool(cfg, app, 0, nil)
 
-	loop := sp.newLoop(&Target{}, nil, 0, false, nil)
+	loop := sp.newLoop(scrapeLoopOptions{
+		target: &Target{},
+	})
 	appl, ok := loop.(*scrapeLoop)
-	if !ok {
-		t.Fatalf("Expected scrapeLoop but got %T", loop)
-	}
+	testutil.Assert(t, ok, "Expected scrapeLoop but got %T", loop)
+
 	wrapped := appl.appender()
 
 	tl, ok := wrapped.(*timeLimitAppender)
-	if !ok {
-		t.Fatalf("Expected timeLimitAppender but got %T", wrapped)
-	}
-	if _, ok := tl.Appender.(nopAppender); !ok {
-		t.Fatalf("Expected base appender but got %T", tl.Appender)
-	}
+	testutil.Assert(t, ok, "Expected timeLimitAppender but got %T", wrapped)
 
-	loop = sp.newLoop(&Target{}, nil, 100, false, nil)
+	_, ok = tl.Appender.(nopAppender)
+	testutil.Assert(t, ok, "Expected base appender but got %T", tl.Appender)
+
+	loop = sp.newLoop(scrapeLoopOptions{
+		target: &Target{},
+		limit:  100,
+	})
 	appl, ok = loop.(*scrapeLoop)
-	if !ok {
-		t.Fatalf("Expected scrapeLoop but got %T", loop)
-	}
+	testutil.Assert(t, ok, "Expected scrapeLoop but got %T", loop)
+
 	wrapped = appl.appender()
 
 	sl, ok := wrapped.(*limitAppender)
-	if !ok {
-		t.Fatalf("Expected limitAppender but got %T", wrapped)
-	}
+	testutil.Assert(t, ok, "Expected limitAppender but got %T", wrapped)
+
 	tl, ok = sl.Appender.(*timeLimitAppender)
-	if !ok {
-		t.Fatalf("Expected limitAppender but got %T", sl.Appender)
-	}
-	if _, ok := tl.Appender.(nopAppender); !ok {
-		t.Fatalf("Expected base appender but got %T", tl.Appender)
-	}
+	testutil.Assert(t, ok, "Expected limitAppender but got %T", sl.Appender)
+
+	_, ok = tl.Appender.(nopAppender)
+	testutil.Assert(t, ok, "Expected base appender but got %T", tl.Appender)
 }
 
 func TestScrapePoolRaces(t *testing.T) {
@@ -350,7 +332,7 @@ func TestScrapePoolRaces(t *testing.T) {
 	newConfig := func() *config.ScrapeConfig {
 		return &config.ScrapeConfig{ScrapeInterval: interval, ScrapeTimeout: timeout}
 	}
-	sp := newScrapePool(newConfig(), &nopAppendable{}, nil)
+	sp, _ := newScrapePool(newConfig(), &nopAppendable{}, 0, nil)
 	tgts := []*targetgroup.Group{
 		{
 			Targets: []model.LabelSet{
@@ -370,12 +352,9 @@ func TestScrapePoolRaces(t *testing.T) {
 	active := sp.ActiveTargets()
 	dropped := sp.DroppedTargets()
 	expectedActive, expectedDropped := len(tgts[0].Targets), 0
-	if len(sp.ActiveTargets()) != expectedActive {
-		t.Fatalf("Invalid number of active targets: expected %v, got %v", expectedActive, len(active))
-	}
-	if len(dropped) != expectedDropped {
-		t.Fatalf("Invalid number of dropped targets: expected %v, got %v", expectedDropped, len(dropped))
-	}
+
+	testutil.Equals(t, expectedActive, len(active), "Invalid number of active targets")
+	testutil.Equals(t, expectedDropped, len(dropped), "Invalid number of dropped targets")
 
 	for i := 0; i < 20; i++ {
 		time.Sleep(time.Duration(10 * time.Millisecond))
@@ -392,7 +371,8 @@ func TestScrapeLoopStopBeforeRun(t *testing.T) {
 		nil, nil,
 		nopMutator,
 		nopMutator,
-		nil, nil,
+		nil, nil, 0,
+		true,
 	)
 
 	// The scrape pool synchronizes on stopping scrape loops. However, new scrape
@@ -456,6 +436,8 @@ func TestScrapeLoopStop(t *testing.T) {
 		nopMutator,
 		app,
 		nil,
+		0,
+		true,
 	)
 
 	// Terminate loop after 2 scrapes.
@@ -481,15 +463,15 @@ func TestScrapeLoopStop(t *testing.T) {
 		t.Fatalf("Scrape wasn't stopped.")
 	}
 
-	// We expected 1 actual sample for each scrape plus 4 for report samples.
+	// We expected 1 actual sample for each scrape plus 5 for report samples.
 	// At least 2 scrapes were made, plus the final stale markers.
-	if len(appender.result) < 5*3 || len(appender.result)%5 != 0 {
-		t.Fatalf("Expected at least 3 scrapes with 4 samples each, got %d samples", len(appender.result))
+	if len(appender.result) < 6*3 || len(appender.result)%6 != 0 {
+		t.Fatalf("Expected at least 3 scrapes with 6 samples each, got %d samples", len(appender.result))
 	}
 	// All samples in a scrape must have the same timestamp.
 	var ts int64
 	for i, s := range appender.result {
-		if i%5 == 0 {
+		if i%6 == 0 {
 			ts = s.t
 		} else if s.t != ts {
 			t.Fatalf("Unexpected multiple timestamps within single scrape")
@@ -521,6 +503,8 @@ func TestScrapeLoopRun(t *testing.T) {
 		nopMutator,
 		app,
 		nil,
+		0,
+		true,
 	)
 
 	// The loop must terminate during the initial offset if the context
@@ -539,12 +523,12 @@ func TestScrapeLoopRun(t *testing.T) {
 	select {
 	case <-signal:
 	case <-time.After(5 * time.Second):
-		t.Fatalf("Cancelation during initial offset failed")
+		t.Fatalf("Cancellation during initial offset failed")
 	case err := <-errc:
 		t.Fatalf("Unexpected error: %s", err)
 	}
 
-	// The provided timeout must cause cancelation of the context passed down to the
+	// The provided timeout must cause cancellation of the context passed down to the
 	// scraper. The scraper has to respect the context.
 	scraper.offsetDur = 0
 
@@ -566,6 +550,8 @@ func TestScrapeLoopRun(t *testing.T) {
 		nopMutator,
 		app,
 		nil,
+		0,
+		true,
 	)
 
 	go func() {
@@ -594,7 +580,7 @@ func TestScrapeLoopRun(t *testing.T) {
 	case err := <-errc:
 		t.Fatalf("Unexpected error: %s", err)
 	case <-time.After(3 * time.Second):
-		t.Fatalf("Loop did not terminate on context cancelation")
+		t.Fatalf("Loop did not terminate on context cancellation")
 	}
 }
 
@@ -614,10 +600,12 @@ func TestScrapeLoopMetadata(t *testing.T) {
 		nopMutator,
 		func() storage.Appender { return nopAppender{} },
 		cache,
+		0,
+		true,
 	)
 	defer cancel()
 
-	total, _, err := sl.append([]byte(`# TYPE test_metric counter
+	total, _, _, err := sl.append([]byte(`# TYPE test_metric counter
 # HELP test_metric some help text
 # UNIT test_metric metric
 test_metric 1
@@ -627,25 +615,59 @@ test_metric 1
 	testutil.Ok(t, err)
 	testutil.Equals(t, 1, total)
 
-	md, ok := cache.getMetadata("test_metric")
+	md, ok := cache.GetMetadata("test_metric")
 	testutil.Assert(t, ok, "expected metadata to be present")
 	testutil.Assert(t, textparse.MetricTypeCounter == md.Type, "unexpected metric type")
 	testutil.Equals(t, "some help text", md.Help)
 	testutil.Equals(t, "metric", md.Unit)
 
-	md, ok = cache.getMetadata("test_metric_no_help")
+	md, ok = cache.GetMetadata("test_metric_no_help")
 	testutil.Assert(t, ok, "expected metadata to be present")
 	testutil.Assert(t, textparse.MetricTypeGauge == md.Type, "unexpected metric type")
 	testutil.Equals(t, "", md.Help)
 	testutil.Equals(t, "", md.Unit)
 
-	md, ok = cache.getMetadata("test_metric_no_type")
+	md, ok = cache.GetMetadata("test_metric_no_type")
 	testutil.Assert(t, ok, "expected metadata to be present")
 	testutil.Assert(t, textparse.MetricTypeUnknown == md.Type, "unexpected metric type")
 	testutil.Equals(t, "other help text", md.Help)
 	testutil.Equals(t, "", md.Unit)
 }
 
+func TestScrapeLoopSeriesAdded(t *testing.T) {
+	// Need a full storage for correct Add/AddFast semantics.
+	s := teststorage.New(t)
+	defer s.Close()
+
+	app, err := s.Appender()
+	testutil.Ok(t, err)
+
+	ctx, cancel := context.WithCancel(context.Background())
+	sl := newScrapeLoop(ctx,
+		&testScraper{},
+		nil, nil,
+		nopMutator,
+		nopMutator,
+		func() storage.Appender { return app },
+		nil,
+		0,
+		true,
+	)
+	defer cancel()
+
+	total, added, seriesAdded, err := sl.append([]byte("test_metric 1\n"), "", time.Time{})
+	testutil.Ok(t, err)
+	testutil.Equals(t, 1, total)
+	testutil.Equals(t, 1, added)
+	testutil.Equals(t, 1, seriesAdded)
+
+	total, added, seriesAdded, err = sl.append([]byte("test_metric 1\n"), "", time.Time{})
+	testutil.Ok(t, err)
+	testutil.Equals(t, 1, total)
+	testutil.Equals(t, 1, added)
+	testutil.Equals(t, 0, seriesAdded)
+}
+
 func TestScrapeLoopRunCreatesStaleMarkersOnFailedScrape(t *testing.T) {
 	appender := &collectResultAppender{}
 	var (
@@ -663,6 +685,8 @@ func TestScrapeLoopRunCreatesStaleMarkersOnFailedScrape(t *testing.T) {
 		nopMutator,
 		app,
 		nil,
+		0,
+		true,
 	)
 	// Succeed once, several failures, then stop.
 	numScrapes := 0
@@ -676,7 +700,7 @@ func TestScrapeLoopRunCreatesStaleMarkersOnFailedScrape(t *testing.T) {
 		} else if numScrapes == 5 {
 			cancel()
 		}
-		return fmt.Errorf("scrape failed")
+		return errors.New("scrape failed")
 	}
 
 	go func() {
@@ -690,17 +714,12 @@ func TestScrapeLoopRunCreatesStaleMarkersOnFailedScrape(t *testing.T) {
 		t.Fatalf("Scrape wasn't stopped.")
 	}
 
-	// 1 successfully scraped sample, 1 stale marker after first fail, 4 report samples for
+	// 1 successfully scraped sample, 1 stale marker after first fail, 5 report samples for
 	// each scrape successful or not.
-	if len(appender.result) != 22 {
-		t.Fatalf("Appended samples not as expected. Wanted: %d samples Got: %d", 22, len(appender.result))
-	}
-	if appender.result[0].v != 42.0 {
-		t.Fatalf("Appended first sample not as expected. Wanted: %f Got: %f", appender.result[0].v, 42.0)
-	}
-	if !value.IsStaleNaN(appender.result[5].v) {
-		t.Fatalf("Appended second sample not as expected. Wanted: stale NaN Got: %x", math.Float64bits(appender.result[5].v))
-	}
+	testutil.Equals(t, 27, len(appender.result), "Appended samples not as expected")
+	testutil.Equals(t, 42.0, appender.result[0].v, "Appended first sample not as expected")
+	testutil.Assert(t, value.IsStaleNaN(appender.result[6].v),
+		"Appended second sample not as expected. Wanted: stale NaN Got: %x", math.Float64bits(appender.result[6].v))
 }
 
 func TestScrapeLoopRunCreatesStaleMarkersOnParseFailure(t *testing.T) {
@@ -721,6 +740,8 @@ func TestScrapeLoopRunCreatesStaleMarkersOnParseFailure(t *testing.T) {
 		nopMutator,
 		app,
 		nil,
+		0,
+		true,
 	)
 
 	// Succeed once, several failures, then stop.
@@ -736,6 +757,85 @@ func TestScrapeLoopRunCreatesStaleMarkersOnParseFailure(t *testing.T) {
 		} else if numScrapes == 3 {
 			cancel()
 		}
+		return errors.New("scrape failed")
+	}
+
+	go func() {
+		sl.run(10*time.Millisecond, time.Hour, nil)
+		signal <- struct{}{}
+	}()
+
+	select {
+	case <-signal:
+	case <-time.After(5 * time.Second):
+		t.Fatalf("Scrape wasn't stopped.")
+	}
+
+	// 1 successfully scraped sample, 1 stale marker after first fail, 5 report samples for
+	// each scrape successful or not.
+	testutil.Equals(t, 17, len(appender.result), "Appended samples not as expected")
+	testutil.Equals(t, 42.0, appender.result[0].v, "Appended first sample not as expected")
+	testutil.Assert(t, value.IsStaleNaN(appender.result[6].v),
+		"Appended second sample not as expected. Wanted: stale NaN Got: %x", math.Float64bits(appender.result[6].v))
+}
+
+func TestScrapeLoopCache(t *testing.T) {
+	s := teststorage.New(t)
+	defer s.Close()
+
+	sapp, err := s.Appender()
+	testutil.Ok(t, err)
+
+	appender := &collectResultAppender{next: sapp}
+	var (
+		signal  = make(chan struct{})
+		scraper = &testScraper{}
+		app     = func() storage.Appender { return appender }
+	)
+	defer close(signal)
+
+	ctx, cancel := context.WithCancel(context.Background())
+	sl := newScrapeLoop(ctx,
+		scraper,
+		nil, nil,
+		nopMutator,
+		nopMutator,
+		app,
+		nil,
+		0,
+		true,
+	)
+
+	numScrapes := 0
+
+	scraper.scrapeFunc = func(ctx context.Context, w io.Writer) error {
+		if numScrapes == 1 || numScrapes == 2 {
+			if _, ok := sl.cache.series["metric_a"]; !ok {
+				t.Errorf("metric_a missing from cache after scrape %d", numScrapes)
+			}
+			if _, ok := sl.cache.series["metric_b"]; !ok {
+				t.Errorf("metric_b missing from cache after scrape %d", numScrapes)
+			}
+		} else if numScrapes == 3 {
+			if _, ok := sl.cache.series["metric_a"]; !ok {
+				t.Errorf("metric_a missing from cache after scrape %d", numScrapes)
+			}
+			if _, ok := sl.cache.series["metric_b"]; ok {
+				t.Errorf("metric_b present in cache after scrape %d", numScrapes)
+			}
+		}
+
+		numScrapes++
+
+		if numScrapes == 1 {
+			w.Write([]byte("metric_a 42\nmetric_b 43\n"))
+			return nil
+		} else if numScrapes == 3 {
+			w.Write([]byte("metric_a 44\n"))
+			return nil
+		} else if numScrapes == 4 {
+			cancel()
+		}
 		return fmt.Errorf("scrape failed")
 	}
 
@@ -750,16 +850,69 @@ func TestScrapeLoopRunCreatesStaleMarkersOnParseFailure(t *testing.T) {
 		t.Fatalf("Scrape wasn't stopped.")
 	}
 
-	// 1 successfully scraped sample, 1 stale marker after first fail, 4 report samples for
+	// 1 successfully scraped sample, 1 stale marker after first fail, 5 report samples for
 	// each scrape successful or not.
-	if len(appender.result) != 14 {
-		t.Fatalf("Appended samples not as expected. Wanted: %d samples Got: %d", 22, len(appender.result))
+	if len(appender.result) != 26 {
+		t.Fatalf("Appended samples not as expected. Wanted: %d samples Got: %d", 26, len(appender.result))
 	}
-	if appender.result[0].v != 42.0 {
-		t.Fatalf("Appended first sample not as expected. Wanted: %f Got: %f", appender.result[0].v, 42.0)
+}
+
+func TestScrapeLoopCacheMemoryExhaustionProtection(t *testing.T) {
+	s := teststorage.New(t)
+	defer s.Close()
+
+	sapp, err := s.Appender()
+	testutil.Ok(t, err)
+
+	appender := &collectResultAppender{next: sapp}
+	var (
+		signal  = make(chan struct{})
+		scraper = &testScraper{}
+		app     = func() storage.Appender { return appender }
+	)
+	defer close(signal)
+
+	ctx, cancel := context.WithCancel(context.Background())
+	sl := newScrapeLoop(ctx,
+		scraper,
+		nil, nil,
+		nopMutator,
+		nopMutator,
+		app,
+		nil,
+		0,
+		true,
+	)
+
+	numScrapes := 0
+
+	scraper.scrapeFunc = func(ctx context.Context, w io.Writer) error {
+		numScrapes++
+		if numScrapes < 5 {
+			s := ""
+			for i := 0; i < 500; i++ {
+				s = fmt.Sprintf("%smetric_%d_%d 42\n", s, i, numScrapes)
+			}
+			w.Write([]byte(fmt.Sprintf(s + "&")))
+		} else {
+			cancel()
+		}
+		return nil
 	}
-	if !value.IsStaleNaN(appender.result[5].v) {
-		t.Fatalf("Appended second sample not as expected. Wanted: stale NaN Got: %x", math.Float64bits(appender.result[5].v))
+
+	go func() {
+		sl.run(10*time.Millisecond, time.Hour, nil)
+		signal <- struct{}{}
+	}()
+
+	select {
+	case <-signal:
+	case <-time.After(5 * time.Second):
+		t.Fatalf("Scrape wasn't stopped.")
+	}
+
+	if len(sl.cache.series) > 2000 {
+		t.Fatalf("More than 2000 series cached. Got: %d", len(sl.cache.series))
 	}
 }
 
@@ -782,6 +935,15 @@ func TestScrapeLoopAppend(t *testing.T) {
 			discoveryLabels: []string{"n", "2"},
 			expLset:         labels.FromStrings("__name__", "metric", "exported_n", "1", "n", "2"),
 			expValue:        0,
+		}, {
+			// When "honor_labels" is not set
+			// exported label from discovery don't get overwritten
+			title:           "Label name collision",
+			honorLabels:     false,
+			scrapeLabels:    `metric 0`,
+			discoveryLabels: []string{"n", "2", "exported_n", "2"},
+			expLset:         labels.FromStrings("__name__", "metric", "n", "2", "exported_n", "2"),
+			expValue:        0,
 		}, {
 			// Labels with no value need to be removed as these should not be ingested.
 			title:           "Delete Empty labels",
@@ -825,14 +987,14 @@ func TestScrapeLoopAppend(t *testing.T) {
 			},
 			func() storage.Appender { return app },
 			nil,
+			0,
+			true,
 		)
 
 		now := time.Now()
 
-		_, _, err := sl.append([]byte(test.scrapeLabels), "", now)
-		if err != nil {
-			t.Fatalf("Unexpected append error: %s", err)
-		}
+		_, _, _, err := sl.append([]byte(test.scrapeLabels), "", now)
+		testutil.Ok(t, err)
 
 		expected := []sample{
 			{
@@ -864,33 +1026,32 @@ func TestScrapeLoopAppendSampleLimit(t *testing.T) {
 		nopMutator,
 		func() storage.Appender { return app },
 		nil,
+		0,
+		true,
 	)
 
 	// Get the value of the Counter before performing the append.
 	beforeMetric := dto.Metric{}
 	err := targetScrapeSampleLimit.Write(&beforeMetric)
-	if err != nil {
-		t.Fatal(err)
-	}
+	testutil.Ok(t, err)
+
 	beforeMetricValue := beforeMetric.GetCounter().GetValue()
 
 	now := time.Now()
-	_, _, err = sl.append([]byte("metric_a 1\nmetric_b 1\nmetric_c 1\n"), "", now)
+	_, _, _, err = sl.append([]byte("metric_a 1\nmetric_b 1\nmetric_c 1\n"), "", now)
 	if err != errSampleLimit {
 		t.Fatalf("Did not see expected sample limit error: %s", err)
 	}
 
-	// Check that the Counter has been incremented a simgle time for the scrape,
+	// Check that the Counter has been incremented a single time for the scrape,
 	// not multiple times for each sample.
 	metric := dto.Metric{}
 	err = targetScrapeSampleLimit.Write(&metric)
-	if err != nil {
-		t.Fatal(err)
-	}
+	testutil.Ok(t, err)
+
 	value := metric.GetCounter().GetValue()
-	if (value - beforeMetricValue) != 1 {
-		t.Fatalf("Unexpected change of sample limit metric: %f", (value - beforeMetricValue))
-	}
+	change := value - beforeMetricValue
+	testutil.Assert(t, change == 1, "Unexpected change of sample limit metric: %f", change)
 
 	// And verify that we got the samples that fit under the limit.
 	want := []sample{
@@ -900,22 +1061,19 @@ func TestScrapeLoopAppendSampleLimit(t *testing.T) {
 			v:      1,
 		},
 	}
-	if !reflect.DeepEqual(want, resApp.result) {
-		t.Fatalf("Appended samples not as expected. Wanted: %+v Got: %+v", want, resApp.result)
-	}
+	testutil.Equals(t, want, resApp.result, "Appended samples not as expected")
 }
 
 func TestScrapeLoop_ChangingMetricString(t *testing.T) {
 	// This is a regression test for the scrape loop cache not properly maintaining
 	// IDs when the string representation of a metric changes across a scrape. Thus
 	// we use a real storage appender here.
-	s := testutil.NewStorage(t)
+	s := teststorage.New(t)
 	defer s.Close()
 
 	app, err := s.Appender()
-	if err != nil {
-		t.Error(err)
-	}
+	testutil.Ok(t, err)
+
 	capp := &collectResultAppender{next: app}
 
 	sl := newScrapeLoop(context.Background(),
@@ -924,17 +1082,16 @@ func TestScrapeLoop_ChangingMetricString(t *testing.T) {
 		nopMutator,
 		func() storage.Appender { return capp },
 		nil,
+		0,
+		true,
 	)
 
 	now := time.Now()
-	_, _, err = sl.append([]byte(`metric_a{a="1",b="1"} 1`), "", now)
-	if err != nil {
-		t.Fatalf("Unexpected append error: %s", err)
-	}
-	_, _, err = sl.append([]byte(`metric_a{b="1",a="1"} 2`), "", now.Add(time.Minute))
-	if err != nil {
-		t.Fatalf("Unexpected append error: %s", err)
-	}
+	_, _, _, err = sl.append([]byte(`metric_a{a="1",b="1"} 1`), "", now)
+	testutil.Ok(t, err)
+
+	_, _, _, err = sl.append([]byte(`metric_a{b="1",a="1"} 2`), "", now.Add(time.Minute))
+	testutil.Ok(t, err)
 
 	// DeepEqual will report NaNs as being different, so replace with a different value.
 	want := []sample{
@@ -949,9 +1106,7 @@ func TestScrapeLoop_ChangingMetricString(t *testing.T) {
 			v:      2,
 		},
 	}
-	if !reflect.DeepEqual(want, capp.result) {
-		t.Fatalf("Appended samples not as expected. Wanted: %+v Got: %+v", want, capp.result)
-	}
+	testutil.Equals(t, want, capp.result, "Appended samples not as expected")
 }
 
 func TestScrapeLoopAppendStaleness(t *testing.T) {
@@ -963,22 +1118,19 @@ func TestScrapeLoopAppendStaleness(t *testing.T) {
 		nopMutator,
 		func() storage.Appender { return app },
 		nil,
+		0,
+		true,
 	)
 
 	now := time.Now()
-	_, _, err := sl.append([]byte("metric_a 1\n"), "", now)
-	if err != nil {
-		t.Fatalf("Unexpected append error: %s", err)
-	}
-	_, _, err = sl.append([]byte(""), "", now.Add(time.Second))
-	if err != nil {
-		t.Fatalf("Unexpected append error: %s", err)
-	}
+	_, _, _, err := sl.append([]byte("metric_a 1\n"), "", now)
+	testutil.Ok(t, err)
+
+	_, _, _, err = sl.append([]byte(""), "", now.Add(time.Second))
+	testutil.Ok(t, err)
 
 	ingestedNaN := math.Float64bits(app.result[1].v)
-	if ingestedNaN != value.StaleNaN {
-		t.Fatalf("Appended stale sample wasn't as expected. Wanted: %x Got: %x", value.StaleNaN, ingestedNaN)
-	}
+	testutil.Equals(t, value.StaleNaN, ingestedNaN, "Appended stale sample wasn't as expected")
 
 	// DeepEqual will report NaNs as being different, so replace with a different value.
 	app.result[1].v = 42
@@ -994,10 +1146,7 @@ func TestScrapeLoopAppendStaleness(t *testing.T) {
 			v:      42,
 		},
 	}
-	if !reflect.DeepEqual(want, app.result) {
-		t.Fatalf("Appended samples not as expected. Wanted: %+v Got: %+v", want, app.result)
-	}
-
+	testutil.Equals(t, want, app.result, "Appended samples not as expected")
 }
 
 func TestScrapeLoopAppendNoStalenessIfTimestamp(t *testing.T) {
@@ -1008,17 +1157,16 @@ func TestScrapeLoopAppendNoStalenessIfTimestamp(t *testing.T) {
 		nopMutator,
 		func() storage.Appender { return app },
 		nil,
+		0,
+		true,
 	)
 
 	now := time.Now()
-	_, _, err := sl.append([]byte("metric_a 1 1000\n"), "", now)
-	if err != nil {
-		t.Fatalf("Unexpected append error: %s", err)
-	}
-	_, _, err = sl.append([]byte(""), "", now.Add(time.Second))
-	if err != nil {
-		t.Fatalf("Unexpected append error: %s", err)
-	}
+	_, _, _, err := sl.append([]byte("metric_a 1 1000\n"), "", now)
+	testutil.Ok(t, err)
+
+	_, _, _, err = sl.append([]byte(""), "", now.Add(time.Second))
+	testutil.Ok(t, err)
 
 	want := []sample{
 		{
@@ -1027,9 +1175,7 @@ func TestScrapeLoopAppendNoStalenessIfTimestamp(t *testing.T) {
 			v:      1,
 		},
 	}
-	if !reflect.DeepEqual(want, app.result) {
-		t.Fatalf("Appended samples not as expected. Wanted: %+v Got: %+v", want, app.result)
-	}
+	testutil.Equals(t, want, app.result, "Appended samples not as expected")
 }
 
 func TestScrapeLoopRunReportsTargetDownOnScrapeError(t *testing.T) {
@@ -1047,18 +1193,17 @@ func TestScrapeLoopRunReportsTargetDownOnScrapeError(t *testing.T) {
 		nopMutator,
 		app,
 		nil,
+		0,
+		true,
 	)
 
 	scraper.scrapeFunc = func(ctx context.Context, w io.Writer) error {
 		cancel()
-		return fmt.Errorf("scrape failed")
+		return errors.New("scrape failed")
 	}
 
 	sl.run(10*time.Millisecond, time.Hour, nil)
-
-	if appender.result[0].v != 0 {
-		t.Fatalf("bad 'up' value; want 0, got %v", appender.result[0].v)
-	}
+	testutil.Equals(t, 0.0, appender.result[0].v, "bad 'up' value")
 }
 
 func TestScrapeLoopRunReportsTargetDownOnInvalidUTF8(t *testing.T) {
@@ -1076,6 +1221,8 @@ func TestScrapeLoopRunReportsTargetDownOnInvalidUTF8(t *testing.T) {
 		nopMutator,
 		app,
 		nil,
+		0,
+		true,
 	)
 
 	scraper.scrapeFunc = func(ctx context.Context, w io.Writer) error {
@@ -1085,10 +1232,7 @@ func TestScrapeLoopRunReportsTargetDownOnInvalidUTF8(t *testing.T) {
 	}
 
 	sl.run(10*time.Millisecond, time.Hour, nil)
-
-	if appender.result[0].v != 0 {
-		t.Fatalf("bad 'up' value; want 0, got %v", appender.result[0].v)
-	}
+	testutil.Equals(t, 0.0, appender.result[0].v, "bad 'up' value")
 }
 
 type errorAppender struct {
@@ -1122,13 +1266,14 @@ func TestScrapeLoopAppendGracefullyIfAmendOrOutOfOrderOrOutOfBounds(t *testing.T
 		nopMutator,
 		func() storage.Appender { return app },
 		nil,
+		0,
+		true,
 	)
 
 	now := time.Unix(1, 0)
-	_, _, err := sl.append([]byte("out_of_order 1\namend 1\nnormal 1\nout_of_bounds 1\n"), "", now)
-	if err != nil {
-		t.Fatalf("Unexpected append error: %s", err)
-	}
+	total, added, seriesAdded, err := sl.append([]byte("out_of_order 1\namend 1\nnormal 1\nout_of_bounds 1\n"), "", now)
+	testutil.Ok(t, err)
+
 	want := []sample{
 		{
 			metric: labels.FromStrings(model.MetricNameLabel, "normal"),
@@ -1136,9 +1281,10 @@ func TestScrapeLoopAppendGracefullyIfAmendOrOutOfOrderOrOutOfBounds(t *testing.T
 			v:      1,
 		},
 	}
-	if !reflect.DeepEqual(want, app.result) {
-		t.Fatalf("Appended samples not as expected. Wanted: %+v Got: %+v", want, app.result)
-	}
+	testutil.Equals(t, want, app.result, "Appended samples not as expected")
+	testutil.Equals(t, 4, total)
+	testutil.Equals(t, 1, added)
+	testutil.Equals(t, 1, seriesAdded)
 }
 
 func TestScrapeLoopOutOfBoundsTimeError(t *testing.T) {
@@ -1155,22 +1301,17 @@ func TestScrapeLoopOutOfBoundsTimeError(t *testing.T) {
 			}
 		},
 		nil,
+		0,
+		true,
 	)
 
 	now := time.Now().Add(20 * time.Minute)
-	total, added, err := sl.append([]byte("normal 1\n"), "", now)
-	if total != 1 {
-		t.Error("expected 1 metric")
-		return
-	}
+	total, added, seriesAdded, err := sl.append([]byte("normal 1\n"), "", now)
+	testutil.Equals(t, 1, total)
+	testutil.Equals(t, 0, added)
+	testutil.Equals(t, 0, seriesAdded)
 
-	if added != 0 {
-		t.Error("no metric should be added")
-	}
-
-	if err != nil {
-		t.Errorf("expect no error, got %s", err.Error())
-	}
+	testutil.Ok(t, err)
 }
 
 func TestTargetScraperScrapeOK(t *testing.T) {
@@ -1215,11 +1356,9 @@ func TestTargetScraperScrapeOK(t *testing.T) {
 	var buf bytes.Buffer
 
 	contentType, err := ts.scrape(context.Background(), &buf)
-	if err != nil {
-		t.Fatalf("Unexpected scrape error: %s", err)
-	}
-	require.Equal(t, "text/plain; version=0.0.4", contentType)
-	require.Equal(t, "metric_a 1\nmetric_b 2\n", buf.String())
+	testutil.Ok(t, err)
+	testutil.Equals(t, "text/plain; version=0.0.4", contentType)
+	testutil.Equals(t, "metric_a 1\nmetric_b 2\n", buf.String())
 }
 
 func TestTargetScrapeScrapeCancel(t *testing.T) {
@@ -1258,9 +1397,9 @@ func TestTargetScrapeScrapeCancel(t *testing.T) {
 	go func() {
 		_, err := ts.scrape(ctx, ioutil.Discard)
 		if err == nil {
-			errc <- fmt.Errorf("Expected error but got nil")
+			errc <- errors.New("Expected error but got nil")
 		} else if ctx.Err() != context.Canceled {
-			errc <- fmt.Errorf("Expected context cancelation error but got: %s", ctx.Err())
+			errc <- errors.Errorf("Expected context cancellation error but got: %s", ctx.Err())
 		}
 		close(errc)
 	}()
@@ -1269,9 +1408,7 @@ func TestTargetScrapeScrapeCancel(t *testing.T) {
 	case <-time.After(5 * time.Second):
 		t.Fatalf("Scrape function did not return unexpectedly")
 	case err := <-errc:
-		if err != nil {
-			t.Fatalf(err.Error())
-		}
+		testutil.Ok(t, err)
 	}
 	// If this is closed in a defer above the function the test server
 	// doesn't terminate and the test doesn't complete.
@@ -1301,9 +1438,8 @@ func TestTargetScrapeScrapeNotFound(t *testing.T) {
 		client: http.DefaultClient,
 	}
 
-	if _, err := ts.scrape(context.Background(), ioutil.Discard); !strings.Contains(err.Error(), "404") {
-		t.Fatalf("Expected \"404 NotFound\" error but got: %s", err)
-	}
+	_, err = ts.scrape(context.Background(), ioutil.Discard)
+	testutil.Assert(t, strings.Contains(err.Error(), "404"), "Expected \"404 NotFound\" error but got: %s", err)
 }
 
 // testScraper implements the scraper interface and allows setting values
@@ -1319,11 +1455,11 @@ type testScraper struct {
 	scrapeFunc func(context.Context, io.Writer) error
 }
 
-func (ts *testScraper) offset(interval time.Duration) time.Duration {
+func (ts *testScraper) offset(interval time.Duration, jitterSeed uint64) time.Duration {
 	return ts.offsetDur
 }
 
-func (ts *testScraper) report(start time.Time, duration time.Duration, err error) {
+func (ts *testScraper) Report(start time.Time, duration time.Duration, err error) {
 	ts.lastStart = start
 	ts.lastDuration = duration
 	ts.lastError = err
@@ -1335,3 +1471,67 @@ func (ts *testScraper) scrape(ctx context.Context, w io.Writer) (string, error)
 	}
 	return "", ts.scrapeErr
 }
+
+func TestScrapeLoop_RespectTimestamps(t *testing.T) {
+	s := teststorage.New(t)
+	defer s.Close()
+
+	app, err := s.Appender()
+	testutil.Ok(t, err)
+
+	capp := &collectResultAppender{next: app}
+
+	sl := newScrapeLoop(context.Background(),
+		nil, nil, nil,
+		nopMutator,
+		nopMutator,
+		func() storage.Appender { return capp },
+		nil, 0,
+		true,
+	)
+
+	now := time.Now()
+	_, _, _, err = sl.append([]byte(`metric_a{a="1",b="1"} 1 0`), "", now)
+	testutil.Ok(t, err)
+
+	want := []sample{
+		{
+			metric: labels.FromStrings("__name__", "metric_a", "a", "1", "b", "1"),
+			t:      0,
+			v:      1,
+		},
+	}
+	testutil.Equals(t, want, capp.result, "Appended samples not as expected")
+}
+
+func TestScrapeLoop_DiscardTimestamps(t *testing.T) {
+	s := teststorage.New(t)
+	defer s.Close()
+
+	app, err := s.Appender()
+	testutil.Ok(t, err)
+
+	capp := &collectResultAppender{next: app}
+
+	sl := newScrapeLoop(context.Background(),
+		nil, nil, nil,
+		nopMutator,
+		nopMutator,
+		func() storage.Appender { return capp },
+		nil, 0,
+		false,
+	)
+
+	now := time.Now()
+	_, _, _, err = sl.append([]byte(`metric_a{a="1",b="1"} 1 0`), "", now)
+	testutil.Ok(t, err)
+
+	want := []sample{
+		{
+			metric: labels.FromStrings("__name__", "metric_a", "a", "1", "b", "1"),
+			t:      timestamp.FromTime(now),
+			v:      1,
+		},
+	}
+	testutil.Equals(t, want, capp.result, "Appended samples not as expected")
+}
